# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt
import xpart as xp
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_rect_ellipse(test_context):
    np2ctx = test_context.nparray_to_context_array
    ctx2np = test_context.nparray_from_context_array

    aper_rect_ellipse = xt.LimitRectEllipse(_context=test_context,
            max_x=23e-3, max_y=18e-3, a=23e-2, b=23e-2)
    aper_ellipse = xt.LimitRectEllipse(_context=test_context,
                                       a=23e-2, b=23e-2)
    aper_rect = xt.LimitRect(_context=test_context,
                             max_x=23e-3, min_x=-23e-3,
                             max_y=18e-3, min_y=-18e-3)

    XX, YY = np.meshgrid(np.linspace(-30e-3, 30e-3, 100),
                         np.linspace(-30e-3, 30e-3, 100))
    x_part = XX.flatten()
    y_part = XX.flatten()
    part_re = xp.Particles(_context=test_context,
                           x=x_part, y=y_part)
    part_e = part_re.copy()
    part_r = part_re.copy()

    aper_rect_ellipse.track(part_re)
    aper_ellipse.track(part_e)
    aper_rect.track(part_r)

    flag_re = ctx2np(part_re.state)[np.argsort(ctx2np(part_re.particle_id))]
    flag_r = ctx2np(part_r.state)[np.argsort(ctx2np(part_r.particle_id))]
    flag_e = ctx2np(part_e.state)[np.argsort(ctx2np(part_e.particle_id))]

    assert np.all(flag_re == (flag_r & flag_e))


@for_all_test_contexts
def test_aperture_racetrack(test_context):
    part_gen_range = 0.11
    n_part=100000

    aper = xt.LimitRacetrack(_context=test_context,
                             min_x=-5e-2, max_x=10e-2,
                             min_y=-2e-2, max_y=4e-2,
                             a=2e-2, b=1e-2)

    xy_out = np.array([
        [-4.8e-2, 3.7e-2],
        [9.6e-2, 3.7e-2],
        [-4.5e-2, -1.8e-2],
        [9.8e-2, -1.8e-2],
        ])

    xy_in = np.array([
        [-4.2e-2, 3.3e-2],
        [9.4e-2, 3.6e-2],
        [-3.8e-2, -1.8e-2],
        [9.2e-2, -1.8e-2],
        ])

    xy_all = np.concatenate([xy_out, xy_in], axis=0)

    particles = xp.Particles(_context=test_context,
            p0c=6500e9,
            x=xy_all[:, 0],
            y=xy_all[:, 1])

    aper.track(particles)

    part_state = test_context.nparray_from_context_array(particles.state)
    part_id = test_context.nparray_from_context_array(particles.particle_id)

    assert np.all(part_state[part_id<4] == 0)
    assert np.all(part_state[part_id>=4] == 1)


@for_all_test_contexts
def test_aperture_polygon(test_context):
    np2ctx = test_context.nparray_to_context_array
    ctx2np = test_context.nparray_from_context_array

    x_vertices=np.array([1.5, 0.2, -1, -1,  1])*1e-2
    y_vertices=np.array([1.3, 0.5,  1, -1, -1])*1e-2

    aper = xt.LimitPolygon(
                    _context=test_context,
                    x_vertices=np2ctx(x_vertices),
                    y_vertices=np2ctx(y_vertices))

    # Try some particles inside
    parttest = xp.Particles(
                    _context=test_context,
                    p0c=6500e9,
                    x=x_vertices*0.99,
                    y=y_vertices*0.99)
    aper.track(parttest)
    xo.assert_allclose(ctx2np(parttest.state), 1)

    # Try some particles outside
    parttest = xp.Particles(
                    _context=test_context,
                    p0c=6500e9,
                    x=x_vertices*1.01,
                    y=y_vertices*1.01)
    aper.track(parttest)
    xo.assert_allclose(ctx2np(parttest.state), 0)


def test_mad_import():

    from cpymad.madx import Madx

    mad = Madx(stdout=False)

    mad.input("""
        m_circle: marker, apertype="circle", aperture={.2};
        m_ellipse: marker, apertype="ellipse", aperture={.2, .1};
        m_rectangle: marker, apertype="rectangle", aperture={.07, .05};
        m_rectellipse: marker, apertype="rectellipse", aperture={.2, .4, .25, .45};
        m_racetrack: marker, apertype="racetrack", aperture={.6,.4,.2,.1};
        m_octagon: marker, apertype="octagon", aperture={.4, .5, 0.5, 1.};
        m_polygon: marker, apertype="circle", aper_vx= {+5.800e-2,+5.800e-2,-8.800e-2}, aper_vy= {+3.500e-2,-3.500e-2,+0.000e+0};
        beam;
        ss: sequence,l=1;
            m_circle, at=0;
            m_ellipse, at=0.01;
            m_rectangle, at=0.02;
            m_rectellipse, at=0.03;
            m_racetrack, at=0.04;
            m_octagon, at=0.05;
            m_polygon, at=0.06;
        endsequence;

        use,sequence=ss;
        twiss,betx=1,bety=1;
        """
        )

    line = xt.Line.from_madx_sequence(mad.sequence.ss, install_apertures=True)

    apertures = [ee for ee in line.elements if ee.__class__.__name__.startswith('Limit')]

    circ = apertures[0]
    assert circ.__class__.__name__ == 'LimitEllipse'
    xo.assert_allclose(circ.a_squ, .2**2, atol=1e-13, rtol=0)
    xo.assert_allclose(circ.b_squ, .2**2, atol=1e-13, rtol=0)

    ellip = apertures[1]
    assert ellip.__class__.__name__ == 'LimitEllipse'
    xo.assert_allclose(ellip.a_squ, .2**2, atol=1e-13, rtol=0)
    xo.assert_allclose(ellip.b_squ, .1**2, atol=1e-13, rtol=0)

    rect = apertures[2]
    assert rect.__class__.__name__ == 'LimitRect'
    assert rect.min_x == -.07
    assert rect.max_x == +.07
    assert rect.min_y == -.05
    assert rect.max_y == +.05

    rectellip = apertures[3]
    assert rectellip.max_x == .2
    assert rectellip.max_y == .4
    xo.assert_allclose(rectellip.a_squ, .25**2, atol=1e-13, rtol=0)
    xo.assert_allclose(rectellip.b_squ, .45**2, atol=1e-13, rtol=0)

    racetr = apertures[4]
    assert racetr.__class__.__name__ == 'LimitRacetrack'
    assert racetr.min_x == -.6
    assert racetr.max_x == +.6
    assert racetr.min_y == -.4
    assert racetr.max_y == +.4
    assert racetr.a == .2
    assert racetr.b == .1

    octag = apertures[5]
    assert octag.__class__.__name__ == 'LimitPolygon'
    assert octag._xobject.x_vertices[0] == 0.4
    xo.assert_allclose(octag._xobject.y_vertices[0], 0.4*np.tan(0.5), atol=1e-14, rtol=0)
    assert octag._xobject.y_vertices[1] == 0.5
    xo.assert_allclose(octag._xobject.x_vertices[1], 0.5/np.tan(1.), atol=1e-14, rtol=0)

    assert octag._xobject.y_vertices[2] == 0.5
    xo.assert_allclose(octag._xobject.x_vertices[2], -0.5/np.tan(1.), atol=1e-14, rtol=0)
    assert octag._xobject.x_vertices[3] == -0.4
    xo.assert_allclose(octag._xobject.y_vertices[3], 0.4*np.tan(0.5), atol=1e-14, rtol=0)


    assert octag._xobject.x_vertices[4] == -0.4
    xo.assert_allclose(octag._xobject.y_vertices[4], -0.4*np.tan(0.5), atol=1e-14, rtol=0)
    assert octag._xobject.y_vertices[5] == -0.5
    xo.assert_allclose(octag._xobject.x_vertices[5], -0.5/np.tan(1.), atol=1e-14, rtol=0)


    assert octag._xobject.y_vertices[6] == -0.5
    xo.assert_allclose(octag._xobject.x_vertices[6], 0.5/np.tan(1.), atol=1e-14, rtol=0)
    assert octag._xobject.x_vertices[7] == 0.4
    xo.assert_allclose(octag._xobject.y_vertices[7], -0.4*np.tan(0.5), atol=1e-14, rtol=0)

    polyg = apertures[6]
    assert polyg.__class__.__name__ == 'LimitPolygon'
    assert len(polyg._xobject.x_vertices) == 3
    assert len(polyg._xobject.y_vertices) == 3
    assert polyg._xobject.x_vertices[0] == -8.8e-2
    assert polyg._xobject.y_vertices[0] == 0
    assert polyg._xobject.x_vertices[1] == 5.8e-2
    assert polyg._xobject.y_vertices[1] == -3.5e-2


@for_all_test_contexts
def test_longitudinal_rect(test_context):
    np2ctx = test_context.nparray_to_context_array
    ctx2np = test_context.nparray_from_context_array

    aper_rect_longitudinal = xt.LongitudinalLimitRect(_context=test_context,
            min_zeta=-10E-3,max_zeta=20E-3, min_pzeta=-1E-3, max_pzeta = 4E-3)

    coords = np.array([
        [-9E-4, 0.0],
        [15E-3, 0.0],
        [0.0, -8E-4],
        [0.0, 2E-3],
        ])
    particles = xp.Particles(_context=test_context,
                           zeta=coords[:,0], pzeta=coords[:,1])

    aper_rect_longitudinal.track(particles)
    particles.move(_context=xo.ContextCpu())
    assert np.all(particles.state == 1)

    coords = np.array([
        [-11E-3, 0.0],
        [22E-3, 0.0],
        [0.0, -2E-3],
        [0.0, 6E-3],
        ])
    particles = xp.Particles(_context=test_context,
                           zeta=coords[:,0], pzeta=coords[:,1])

    aper_rect_longitudinal.track(particles)
    particles.move(_context=xo.ContextCpu())
    assert np.all(particles.state == -2)

@for_all_test_contexts
def test_aper_tilt(test_context):

    n_part=300000

    particles = xt.Particles(_context=test_context,
            p0c=6500e9,
            x=np.random.uniform(-0.25, 0.25, n_part),
            px = np.zeros(n_part),
            y=np.random.uniform(0, 0.1, n_part),
            py = np.zeros(n_part))

    tilt_deg = 10.
    aper = xt.LimitRect(_context=test_context,
                        min_x=-.1,
                        max_x=.1,
                        min_y=-0.001,
                        max_y=0.001,
                        shift_x=0.08,
                        shift_y=0.04,
                        rot_s_rad=np.deg2rad(tilt_deg))

    aper.track(particles)

    part_id = test_context.nparray_from_context_array(particles.particle_id)
    part_state = test_context.nparray_from_context_array(particles.state)
    part_x = test_context.nparray_from_context_array(particles.x)
    part_y = test_context.nparray_from_context_array(particles.y)

    x_alive = part_x[part_state>0]
    y_alive = part_y[part_state>0]

    assert_allclose = np.testing.assert_allclose
    assert_allclose(np.mean(x_alive), 0.08, rtol=5e-2, atol=0)
    assert_allclose(np.mean(y_alive), 0.04, rtol=5e-2, atol=0)
    slope = np.polyfit(x_alive, y_alive, 1)[0]
    assert_allclose(slope, np.tan(np.deg2rad(tilt_deg)), rtol=5e-2, atol=0)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time

import json
import numpy as np

import xobjects as xo
import xtrack as xt
import xpart as xp

context = xo.ContextCpu()

###############
# Load a line #
###############

fname_line_particles = '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json'

with open(fname_line_particles, 'r') as fid:
    input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])
line.particle_ref = xp.Particles.from_dict(input_data['particle'])

#################
# Build tracker #
#################

line.build_tracker(_context=context)

n_part = 1
particles0 = line.build_particles(x_norm=np.linspace(-1, 1, n_part),
                                nemitt_x=2.5e-6, nemitt_y=2.5e-6)

num_turns = 1000
particles = particles0.copy()
t1 = time.time()
for i_turn in range(num_turns):
    line.track(particles)
t2 = time.time()
print('Time only track: {:.3f} ms/turn'.format((t2 - t1) / num_turns * 1e3))
t_only_track = (t2-t1)

particles = particles0.copy()
t1 = time.time()
for i_turn in range(num_turns):
    line.vars['on_crab1'] = (num_turns - i_turn)/num_turns
    line.track(particles)
t2 = time.time()
print('Overhead track with knob: {:.3f} ms/turn'.format((t2 - t1 - t_only_track) / num_turns * 1e3))

particles = particles0.copy()
t1 = time.time()
for i_turn in range(num_turns):
    line['acfga.4bl1.b1'].knl[0] = 1e-11*i_turn
    line['acfga.4ar1.b1'].ksl[0] = 1e-11*i_turn
    line['acfga.4br1.b1'].ksl[0] = 1e-11*i_turn
    line['acfga.4al1.b1'].ksl[0] = 1e-11*i_turn
    line['acfga.4al1.b1'].knl[0] = 1e-11*i_turn
    line['acfga.4bl1.b1'].ksl[0] = 1e-11*i_turn
    line['acfga.4ar1.b1'].knl[0] = 1e-11*i_turn
    line['acfga.4br1.b1'].knl[0] = 1e-11*i_turn
    line.track(particles)
t2 = time.time()
print('Overhead track with k trim: {:.3f} ms/turn'.format((t2 - t1 - t_only_track) / num_turns * 1e3))


particles = particles0.copy()
t1 = time.time()
for i_turn in range(num_turns):
    line.vars['vrf400'] = (num_turns - i_turn)/num_turns
    line.vars['vrf400'] = (num_turns - i_turn)/num_turns
    line.track(particles)
t2 = time.time()
print('Overhead track with voltage knob: {:.3f} ms/turn'.format((t2 - t1 - t_only_track) / num_turns * 1e3))



"""
Monitor to save particle state for last N turns before respective loss

Author: Philipp Niedermayer
Date: 2023-01-19
"""


import numpy as np

import xobjects as xo
import xtrack as xt

from ..base_element import BeamElement
from ..general import _pkg_root


class LastTurnsData(xo.Struct):

    lost_at_offset = xo.UInt32[:]

    # store with 32bit precission to save memory
    particle_id = xo.UInt32[:]  # TODO: can use xo.Int32[:] after fixing https://github.com/xsuite/xsuite/issues/283
    at_turn = xo.UInt32[:]      # TODO: can use xo.Int32[:] after fixing https://github.com/xsuite/xsuite/issues/283
    x = xo.Float32[:]
    px = xo.Float32[:]
    y = xo.Float32[:]
    py = xo.Float32[:]
    delta = xo.Float32[:]
    zeta = xo.Float32[:]



class LastTurnsMonitor(BeamElement):

    _xofields={
        'particle_id_start': xo.Int64,
        'num_particles': xo.Int64,
        'n_last_turns': xo.Int64,
        'every_n_turns': xo.Int64,
        'data': LastTurnsData,
    }

    behaves_like_drift = True
    allow_loss_refinement = True

    # TODO: find a way to dynamically change what properties are being saved by this monitor
    properties = [field.name for field in LastTurnsData._fields if field.name != 'lost_at_offset']


    _extra_c_sources = [
        _pkg_root.joinpath('monitors/last_turns_monitor.h')
    ]


    def __init__(self, *, n_last_turns=None, num_particles=None, particle_id_range=None, every_n_turns=1, _xobject=None, **kwargs):
        """Monitor to save particle data in last turns before respective particle loss

        The monitor provides the following data as 2D array of shape (num_particles, n_last_turns),
        where the first index corresponds to the particle_id in particle_id_range
        and the second index corresponds to the turn (or every_n_turns) before the respective particle is lost:
        `particle_id`, `at_turn`, `x`, `px`, `y`, `py`, `delta`, `zeta`

        Args:
            n_last_turns (int): Amount of turns to store before particle loss.
            particle_id_range (tuple): Range of particle ids to monitor (start, stop).
            num_particles (int, optional): Number of particles. Equal to passing particle_id_range=(0, num_particles).
            every_n_turns (int, optional): Save only every n-th turn, i.e. turn numbers which are a multiples of this.
                Because `n_last_turns` defines the amount of turns to store (and not the range), the data will cover turn
                numbers up to `n_last_turns*every_n_turns` turns before particle loss.

        Example:
            monitor = LastTurnsMonitor(n_last_turns=5, particle_id_range=(1, 5))
            monitor.at_turn[:,-1]  # last turn before loss of each particle, respectively
            monitor.x[3,-2]  # x coordinate in one but last turn of particle with id 4

        """

        if _xobject is not None:
            super().__init__(_xobject=_xobject)

        else:

            if num_particles is not None and particle_id_range is None:
                particle_id_start = 0
            elif particle_id_range is not None and num_particles is None:
                particle_id_start = particle_id_range[0]
                num_particles = particle_id_range[1] - particle_id_range[0]
            else:
                raise ValueError("Exactly one of `num_particles` or `particle_id_range` parameters must be specified")

            # explicitely init with zeros (instead of size only) to have consistent default values for untouched arrays
            # see also https://github.com/xsuite/xsuite/issues/294
            size = num_particles*n_last_turns
            data = {prop: np.zeros(size) if size > 0 else [] for prop in self.properties} # particle data
            data['lost_at_offset'] = np.zeros(num_particles) if num_particles > 0 else [] # meta data (rolling buffer offset per particle)

            super().__init__(n_last_turns=n_last_turns, particle_id_start=particle_id_start,
                             num_particles=num_particles, every_n_turns=every_n_turns, data=data, **kwargs)

    def __repr__(self):
        return (
            f"{type(self).__qualname__}(n_last_turns={self.n_last_turns}, every_n_turns={self.every_n_turns}, "
            f"particle_id_start={self.particle_id_start}, num_particles={self.num_particles}) at {hex(id(self))}"
        )


    def __getattr__(self, attr):
        if attr in self.properties:
            val = getattr(self.data, attr)
            val = val.to_nparray() # = self.data._buffer.context.nparray_from_context_array(val)
            val = np.reshape(val, (self.num_particles, self.n_last_turns))
            off = self.data.lost_at_offset.to_nparray() + 1
            # correct for rolling buffer offset
            r, c = np.ogrid[:val.shape[0], :val.shape[1]]
            c = (c + off[:, np.newaxis]) % val.shape[1]
            return val[r, c]
        return super().__getattr__(attr)

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return xt.Marker(_context=_context, _buffer=_buffer, _offset=_offset)




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import json

import xtrack as xt
import xpart as xp
import xobjects as xo
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_record_single_table(test_context):
    class TestElementRecord(xo.HybridClass):
        _xofields = {
            '_index': xt.RecordIndex,
            'generated_rr': xo.Float64[:],
            'at_element': xo.Int64[:],
            'at_turn': xo.Int64[:],
            'particle_id': xo.Int64[:]
            }

    extra_src = []

    extra_src.append(r'''
        /*gpufun*/
        void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){

            // Extract the record and record_index
            TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
            RecordIndex record_index = NULL;
            if (record){
                record_index = TestElementRecordData_getp__index(record);
            }

            int64_t n_kicks = TestElementData_get_n_kicks(el);
            //printf("n_kicks %d\n", (int)n_kicks);

            //start_per_particle_block (part0->part)

                for (int64_t i = 0; i < n_kicks; i++) {
                    double rr = 1e-6 * RandomUniform_generate(part);
                    LocalParticle_add_to_px(part, rr);

                    if (record){
                        // Get a slot in the record (this is thread safe)
                        int64_t i_slot = RecordIndex_get_slot(record_index);
                        // The returned slot id is negative if record is NULL or if record is full

                        if (i_slot>=0){
                            TestElementRecordData_set_at_element(record, i_slot,
                                                        LocalParticle_get_at_element(part));
                            TestElementRecordData_set_at_turn(record, i_slot,
                                                        LocalParticle_get_at_turn(part));
                            TestElementRecordData_set_particle_id(record, i_slot,
                                                        LocalParticle_get_particle_id(part));
                            TestElementRecordData_set_generated_rr(record, i_slot, rr);
                        }
                    }
                }


            //end_per_particle_block
        }
        ''')

    class TestElement(xt.BeamElement):
        _xofields={
            'n_kicks': xo.Int64,
            }

        _internal_record_class = TestElementRecord

        _depends_on = [xt.RandomUniform]

        _extra_c_sources = extra_src

    n_kicks0 = 5
    n_kicks1 = 3
    line=xt.Line(elements = [
        TestElement(n_kicks=n_kicks0), TestElement(n_kicks=n_kicks1)])
    line._needs_rng = True
    line.build_tracker(_context=test_context)

    record = line.start_internal_logging_for_elements_of_type(
                                                        TestElement, capacity=10000)

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    line.optimize_for_tracking()
    line.track(part, num_turns=num_turns0)
    line.track(part, num_turns=num_turns1)

    num_recorded = record._index.num_recorded
    num_turns = num_turns0 + num_turns1
    num_particles = len(part.x)
    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())
    assert num_recorded == (num_particles * num_turns * (n_kicks0 + n_kicks1))

    assert np.sum((record.at_element[:num_recorded] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((record.at_element[:num_recorded] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((record.at_turn[:num_recorded] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))

    # Check reached capacity
    record = line.start_internal_logging_for_elements_of_type(
                                                        TestElement, capacity=20)

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    line.track(part, num_turns=num_turns0)
    line.track(part, num_turns=num_turns1)

    num_recorded = record._index.num_recorded
    assert num_recorded == 20


    # Check stop
    record = line.start_internal_logging_for_elements_of_type(
                                                        TestElement, capacity=10000)

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    num_particles = len(part.x)
    line.track(part, num_turns=num_turns0)
    line.stop_internal_logging_for_elements_of_type(TestElement)
    line.track(part, num_turns=num_turns1)

    num_recorded = record._index.num_recorded
    num_turns = num_turns0
    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())
    assert np.all(part.at_turn == num_turns0 + num_turns1)
    assert num_recorded == (num_particles * num_turns
                                            * (n_kicks0 + n_kicks1))

    assert np.sum((record.at_element[:num_recorded] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((record.at_element[:num_recorded] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((record.at_turn[:num_recorded] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))

    # Collective
    n_kicks0 = 5
    n_kicks1 = 3
    elements = [
        TestElement(n_kicks=n_kicks0, _context=test_context), TestElement(n_kicks=n_kicks1)]
    elements[0].iscollective = True
    line=xt.Line(elements=elements)
    line.build_tracker(_context=test_context)
    line._needs_rng = True

    record = line.start_internal_logging_for_elements_of_type(
                                                        TestElement, capacity=10000)

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    line.track(part, num_turns=num_turns0)
    line.stop_internal_logging_for_elements_of_type(TestElement)
    line.track(part, num_turns=num_turns1)

    # Checks
    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())
    num_recorded = record._index.num_recorded
    num_turns = num_turns0
    num_particles = len(part.x)
    assert np.all(part.at_turn == num_turns0 + num_turns1)
    assert num_recorded == (num_particles * num_turns
                                            * (n_kicks0 + n_kicks1))

    assert np.sum((record.at_element[:num_recorded] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((record.at_element[:num_recorded] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((record.at_turn[:num_recorded] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))


@for_all_test_contexts
def test_record_with_twiss(test_context):
    class TestElementRecord(xo.HybridClass):
        _xofields = {
            '_index': xt.RecordIndex,
            'generated_rr': xo.Float64[:],
            'at_element': xo.Int64[:],
            'at_turn': xo.Int64[:],
            'particle_id': xo.Int64[:]
            }

    extra_src = []

    extra_src.append(r'''
        /*gpufun*/
        void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){

            // Extract the record and record_index
            TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
            RecordIndex record_index = NULL;
            if (record){
                record_index = TestElementRecordData_getp__index(record);
            }

            int64_t n_kicks = TestElementData_get_n_kicks(el);
            //printf("n_kicks %d\n", (int)n_kicks);

            //start_per_particle_block (part0->part)

                for (int64_t i = 0; i < n_kicks; i++) {
                    // We don't apply the kick, otherwise the twiss fails
                    double rr = 1e-6 * RandomUniform_generate(part);

                    if (record){
                        // Get a slot in the record (this is thread safe)
                        int64_t i_slot = RecordIndex_get_slot(record_index);
                        // The returned slot id is negative if record is NULL or if record is full

                        if (i_slot>=0){
                            TestElementRecordData_set_at_element(record, i_slot,
                                                        LocalParticle_get_at_element(part));
                            TestElementRecordData_set_at_turn(record, i_slot,
                                                        LocalParticle_get_at_turn(part));
                            TestElementRecordData_set_particle_id(record, i_slot,
                                                        LocalParticle_get_particle_id(part));
                            TestElementRecordData_set_generated_rr(record, i_slot, rr);
                        }
                    }
                }


            //end_per_particle_block
        }
        ''')

    class TestElement(xt.BeamElement):
        _xofields={
            'n_kicks': xo.Int64,
            }

        _internal_record_class = TestElementRecord

        _depends_on = [xt.RandomUniform]

        _extra_c_sources = extra_src

    n_kicks0 = 5
    n_kicks1 = 3

    path_line_particles = xt._pkg_root.parent / 'test_data' / 'hllhc15_noerrors_nobb/line_and_particle.json'

    with open(path_line_particles, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])
    line.particle_ref = xp.Particles.from_dict(input_data['particle'])
    line.insert_element(element=TestElement(n_kicks=n_kicks0), name='test0', at_s=line.get_s_position('ip1'))
    line.insert_element(element=TestElement(n_kicks=n_kicks1), name='test1', at_s=line.get_s_position('ip5'))

    line._needs_rng = True
    line.build_tracker(_context=test_context)
    record = line.start_internal_logging_for_elements_of_type(
                                                        TestElement, capacity=10000)
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=6.5e12)
    io_buffer = line.tracker.io_buffer
    line.twiss(at_s=np.linspace(0, line.get_length(), 500))
    


@for_all_test_contexts
def test_record_multiple_tables(test_context):
    class Table1(xo.HybridClass):
        _xofields = {
            '_index': xt.RecordIndex,
            'particle_x': xo.Float64[:],
            'particle_px': xo.Float64[:],
            'at_element': xo.Int64[:],
            'at_turn': xo.Int64[:],
            'particle_id': xo.Int64[:]
            }

    class Table2(xo.HybridClass):
        _xofields = {
            '_index': xt.RecordIndex,
            'generated_rr': xo.Float64[:],
            'at_element': xo.Int64[:],
            'at_turn': xo.Int64[:],
            'particle_id': xo.Int64[:]
            }

    class TestElementRecord(xo.HybridClass):
        _xofields = {
            'table1': Table1,
            'table2': Table2
            }

    extra_src = []

    extra_src.append(r'''
        /*gpufun*/
        void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){

            // Extract the record and record_index
            TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
            Table1Data table1 = NULL;
            Table2Data table2 = NULL;
            RecordIndex table1_index = NULL;
            RecordIndex table2_index = NULL;
            if (record){
                table1 = TestElementRecordData_getp_table1(record);
                table2 = TestElementRecordData_getp_table2(record);
                table1_index = Table1Data_getp__index(table1);
                table2_index = Table2Data_getp__index(table2);
            }

            int64_t n_kicks = TestElementData_get_n_kicks(el);
            // printf("n_kicks %d\n", (int)n_kicks);

            //start_per_particle_block (part0->part)

                // Record in table1 info about the ingoing particle
                if (record){
                    // Get a slot in table1
                    int64_t i_slot = RecordIndex_get_slot(table1_index);
                    // The returned slot id is negative if record is NULL or if record is full
                    // printf("i_slot %d\n", (int)i_slot);

                    if (i_slot>=0){
                            Table1Data_set_particle_x(table1, i_slot,
                                                        LocalParticle_get_x(part));
                            Table1Data_set_particle_px(table1, i_slot,
                                                        LocalParticle_get_px(part));
                            Table1Data_set_at_element(table1, i_slot,
                                                        LocalParticle_get_at_element(part));
                            Table1Data_set_at_turn(table1, i_slot,
                                                        LocalParticle_get_at_turn(part));
                            Table1Data_set_particle_id(table1, i_slot,
                                                        LocalParticle_get_particle_id(part));
                    }
                }

                for (int64_t i = 0; i < n_kicks; i++) {
                    double rr = 1e-6 * RandomUniform_generate(part);
                    LocalParticle_add_to_px(part, rr);

                    // Record in table2 info about the generated kicks
                    if (record){
                        // Get a slot in table2
                        int64_t i_slot = RecordIndex_get_slot(table2_index);
                        // The returned slot id is negative if record is NULL or if record is full
                        // printf("i_slot %d\n", (int)i_slot);

                        if (i_slot>=0){
                                Table2Data_set_generated_rr(table2, i_slot, rr);
                                Table2Data_set_at_element(table2, i_slot,
                                                            LocalParticle_get_at_element(part));
                                Table2Data_set_at_turn(table2, i_slot,
                                                            LocalParticle_get_at_turn(part));
                                Table2Data_set_particle_id(table2, i_slot,
                                                            LocalParticle_get_particle_id(part));
                        }
                    }
                }

            //end_per_particle_block
        }
        ''')

    class TestElement(xt.BeamElement):
        _xofields={
            'n_kicks': xo.Int64,
            }
        _internal_record_class = TestElementRecord

        _depends_on = [xt.RandomUniform]

        _extra_c_sources = extra_src


    # Checks
    n_kicks0 = 5
    n_kicks1 = 3
    line=xt.Line(elements = [
        TestElement(n_kicks=n_kicks0), TestElement(n_kicks=n_kicks1)])
    line._needs_rng = True
    line.build_tracker(_context=test_context)

    record = line.start_internal_logging_for_elements_of_type(TestElement,
                                capacity={'table1': 10000, 'table2': 10000})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    line.track(part, num_turns=num_turns0)
    line.track(part, num_turns=num_turns1)

    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())

    num_turns = num_turns0 + num_turns1
    num_particles = len(part.x)

    table1 = record.table1
    table2 = record.table2
    num_recorded_tab1 = table1._index.num_recorded
    num_recorded_tab2 = table2._index.num_recorded

    assert num_recorded_tab1 == 2 * (num_particles * num_turns)
    assert num_recorded_tab2 == (num_particles * num_turns * (n_kicks0 + n_kicks1))

    assert np.sum((table1.at_element[:num_recorded_tab1] == 0)) == (num_particles * num_turns)
    assert np.sum((table1.at_element[:num_recorded_tab1] == 1)) == (num_particles * num_turns)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((table1.at_turn[:num_recorded_tab1] == i_turn)) == 2 * num_particles
        assert np.sum((table2.at_turn[:num_recorded_tab2] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))
    # Check reached capacity
    record = line.start_internal_logging_for_elements_of_type(
                                                        TestElement,
                                        capacity={'table1': 20, 'table2': 15})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    line.track(part, num_turns=num_turns0)
    line.track(part, num_turns=num_turns1)

    table1 = record.table1
    table2 = record.table2
    num_recorded_tab1 = table1._index.num_recorded
    num_recorded_tab2 = table2._index.num_recorded

    assert num_recorded_tab1 == 20
    assert num_recorded_tab2 == 15


    # Check stop
    record = line.start_internal_logging_for_elements_of_type(
                                        TestElement,
                                        capacity={'table1': 1000, 'table2': 1000})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    num_particles = len(part.x)

    line.track(part, num_turns=num_turns0)
    line.stop_internal_logging_for_elements_of_type(TestElement)
    line.track(part, num_turns=num_turns1)

    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())

    num_turns = num_turns0

    table1 = record.table1
    table2 = record.table2
    num_recorded_tab1 = table1._index.num_recorded
    num_recorded_tab2 = table2._index.num_recorded

    assert num_recorded_tab1 == 2 * (num_particles * num_turns)
    assert num_recorded_tab2 == (num_particles * num_turns * (n_kicks0 + n_kicks1))

    assert np.sum((table1.at_element[:num_recorded_tab1] == 0)) == (num_particles * num_turns)
    assert np.sum((table1.at_element[:num_recorded_tab1] == 1)) == (num_particles * num_turns)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((table1.at_turn[:num_recorded_tab1] == i_turn)) == 2 * num_particles
        assert np.sum((table2.at_turn[:num_recorded_tab2] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))

    # Collective
    n_kicks0 = 5
    n_kicks1 = 3
    elements = [
        TestElement(n_kicks=n_kicks0, _context=test_context), TestElement(n_kicks=n_kicks1)]
    elements[0].iscollective = True
    line = xt.Line(elements=elements)
    line.build_tracker(_context=test_context)
    line._needs_rng = True

    record = line.start_internal_logging_for_elements_of_type(
                                        TestElement,
                                        capacity={'table1': 1000, 'table2': 1000})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    num_turns0 = 10
    num_turns1 = 3
    line.track(part, num_turns=num_turns0)
    line.stop_internal_logging_for_elements_of_type(TestElement)
    line.track(part, num_turns=num_turns1)

    # Checks
    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())
    num_turns = num_turns0
    num_particles = len(part.x)

    table1 = record.table1
    table2 = record.table2
    num_recorded_tab1 = table1._index.num_recorded
    num_recorded_tab2 = table2._index.num_recorded

    assert num_recorded_tab1 == 2 * (num_particles * num_turns)
    assert num_recorded_tab2 == (num_particles * num_turns * (n_kicks0 + n_kicks1))

    assert np.sum((table1.at_element[:num_recorded_tab1] == 0)) == (num_particles * num_turns)
    assert np.sum((table1.at_element[:num_recorded_tab1] == 1)) == (num_particles * num_turns)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((table1.at_turn[:num_recorded_tab1] == i_turn)) == 2 * num_particles
        assert np.sum((table2.at_turn[:num_recorded_tab2] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))


@for_all_test_contexts
def test_record_standalone_mode(test_context):

    class Table1(xo.HybridClass):
        _xofields = {
            '_index': xt.RecordIndex,
            'particle_x': xo.Float64[:],
            'particle_px': xo.Float64[:],
            'at_element': xo.Int64[:],
            'at_turn': xo.Int64[:],
            'particle_id': xo.Int64[:]
            }

    class Table2(xo.HybridClass):
        _xofields = {
            '_index': xt.RecordIndex,
            'generated_rr': xo.Float64[:],
            'at_element': xo.Int64[:],
            'at_turn': xo.Int64[:],
            'particle_id': xo.Int64[:]
            }

    class TestElementRecord(xo.HybridClass):
        _xofields = {
            'table1': Table1,
            'table2': Table2
            }

    extra_src = []

    extra_src.append(r'''
        /*gpufun*/
        void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){

            // Extract the record and record_index
            TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
            Table1Data table1 = NULL;
            Table2Data table2 = NULL;
            RecordIndex table1_index = NULL;
            RecordIndex table2_index = NULL;
            if (record){
                table1 = TestElementRecordData_getp_table1(record);
                table2 = TestElementRecordData_getp_table2(record);
                table1_index = Table1Data_getp__index(table1);
                table2_index = Table2Data_getp__index(table2);
            }

            int64_t n_kicks = TestElementData_get_n_kicks(el);
            // printf("n_kicks %d\n", (int)n_kicks);

            //start_per_particle_block (part0->part)

                // Record in table1 info about the ingoing particle
                if (record){
                    // Get a slot in table1
                    int64_t i_slot = RecordIndex_get_slot(table1_index);
                    // The returned slot id is negative if record is NULL or if record is full
                    //printf("i_slot %d\n", (int)i_slot);

                    if (i_slot>=0){
                            Table1Data_set_particle_x(table1, i_slot,
                                                        LocalParticle_get_x(part));
                            Table1Data_set_particle_px(table1, i_slot,
                                                        LocalParticle_get_px(part));
                            Table1Data_set_at_element(table1, i_slot,
                                                        LocalParticle_get_at_element(part));
                            Table1Data_set_at_turn(table1, i_slot,
                                                        LocalParticle_get_at_turn(part));
                            Table1Data_set_particle_id(table1, i_slot,
                                                        LocalParticle_get_particle_id(part));
                    }
                }

                for (int64_t i = 0; i < n_kicks; i++) {
                    double rr = 1e-6 * RandomUniform_generate(part);
                    LocalParticle_add_to_px(part, rr);

                    // Record in table2 info about the generated kicks
                    if (record){
                        // Get a slot in table2
                        int64_t i_slot = RecordIndex_get_slot(table2_index);
                        // The returned slot id is negative if record is NULL or if record is full
                        //printf("i_slot %d\n", (int)i_slot);

                        if (i_slot>=0){
                                Table2Data_set_generated_rr(table2, i_slot, rr);
                                Table2Data_set_at_element(table2, i_slot,
                                                            LocalParticle_get_at_element(part));
                                Table2Data_set_at_turn(table2, i_slot,
                                                            LocalParticle_get_at_turn(part));
                                Table2Data_set_particle_id(table2, i_slot,
                                                            LocalParticle_get_particle_id(part));
                        }
                    }
                }

            //end_per_particle_block
        }
        ''')

    class TestElement(xt.BeamElement):
        _xofields={
            'n_kicks': xo.Int64,
            }
        _internal_record_class = TestElementRecord

        _depends_on = [xt.RandomUniform]

        _extra_c_sources = extra_src


    # Checks
    n_kicks0 = 5
    n_kicks1 = 3
    elements = [
        TestElement(_context=test_context, n_kicks=n_kicks0),
        TestElement(_context=test_context, n_kicks=n_kicks1)]

    io_buffer = xt.new_io_buffer(_context=test_context)
    record = xt.start_internal_logging(elements=elements, io_buffer=io_buffer,
                            capacity={'table1': 10000, 'table2': 10000})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    part._init_random_number_generator()
    num_turns0 = 10
    num_turns1 = 3

    for i_turn in range(num_turns0 + num_turns1):
        for ee in elements:
            ee.track(part, increment_at_element=True)
        part.at_element[:] = 0
        part.at_turn += 1

    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())

    num_turns = num_turns0 + num_turns1
    num_particles = len(part.x)

    table1 = record.table1
    table2 = record.table2
    num_recorded_tab1 = table1._index.num_recorded
    num_recorded_tab2 = table2._index.num_recorded

    assert num_recorded_tab1 == 2 * (num_particles * num_turns)
    assert num_recorded_tab2 == (num_particles * num_turns * (n_kicks0 + n_kicks1))

    assert np.sum((table1.at_element[:num_recorded_tab1] == 0)) == (num_particles * num_turns)
    assert np.sum((table1.at_element[:num_recorded_tab1] == 1)) == (num_particles * num_turns)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((table1.at_turn[:num_recorded_tab1] == i_turn)) == 2 * num_particles
        assert np.sum((table2.at_turn[:num_recorded_tab2] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))

    # Check reached capacity
    io_buffer = xt.new_io_buffer(_context=test_context)
    record = xt.start_internal_logging(elements=elements, io_buffer=io_buffer,
                            capacity={'table1': 20, 'table2': 15})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    part._init_random_number_generator()
    num_turns0 = 10
    num_turns1 = 3
    for i_turn in range(num_turns0 + num_turns1):
        for ee in elements:
            ee.track(part, increment_at_element=True)
        part.at_element[:] = 0
        part.at_turn += 1

    table1 = record.table1
    table2 = record.table2
    num_recorded_tab1 = table1._index.num_recorded
    num_recorded_tab2 = table2._index.num_recorded

    assert num_recorded_tab1 == 20
    assert num_recorded_tab2 == 15


    # Check stop
    io_buffer = xt.new_io_buffer(_context=test_context)
    record = xt.start_internal_logging(elements=elements, io_buffer=io_buffer,
                            capacity={'table1': 10000, 'table2': 10000})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    part._init_random_number_generator()
    num_turns0 = 10
    num_turns1 = 3
    num_particles = len(part.x)


    for i_turn in range(num_turns0 + num_turns1):
        if i_turn == num_turns0:
            xt.stop_internal_logging(elements=elements)
        for ee in elements:
            ee.track(part, increment_at_element=True)
        part.at_element[:] = 0
        part.at_turn += 1


    part.move(_context=xo.ContextCpu())
    record.move(_context=xo.ContextCpu())

    num_turns = num_turns0

    table1 = record.table1
    table2 = record.table2
    num_recorded_tab1 = table1._index.num_recorded
    num_recorded_tab2 = table2._index.num_recorded

    assert num_recorded_tab1 == 2 * (num_particles * num_turns)
    assert num_recorded_tab2 == (num_particles * num_turns * (n_kicks0 + n_kicks1))

    assert np.sum((table1.at_element[:num_recorded_tab1] == 0)) == (num_particles * num_turns)
    assert np.sum((table1.at_element[:num_recorded_tab1] == 1)) == (num_particles * num_turns)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((table2.at_element[:num_recorded_tab2] == 1)) == (num_particles * num_turns
                                            * n_kicks1)
    for i_turn in range(num_turns):
        assert np.sum((table1.at_turn[:num_recorded_tab1] == i_turn)) == 2 * num_particles
        assert np.sum((table2.at_turn[:num_recorded_tab2] == i_turn)) == (num_particles
                                                            * (n_kicks0 + n_kicks1))

    # Separate buffers
    io_buffer0 = xt.new_io_buffer(_context=test_context)
    record0 = xt.start_internal_logging(elements=elements[0], io_buffer=io_buffer0,
                            capacity={'table1': 10000, 'table2': 10000})
    io_buffer1 = xt.new_io_buffer(_context=test_context)
    record1 = xt.start_internal_logging(elements=elements[1], io_buffer=io_buffer1,
                            capacity={'table1': 10000, 'table2': 10000})

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1e-3,2e-3,3e-3])
    part._init_random_number_generator()
    num_turns0 = 10
    num_turns1 = 3

    for i_turn in range(num_turns0 + num_turns1):
        for ee in elements:
            ee.track(part, increment_at_element=True)
        part.at_element[:] = 0
        part.at_turn += 1

    part.move(_context=xo.ContextCpu())
    record0.move(_context=xo.ContextCpu())
    record1.move(_context=xo.ContextCpu())

    num_turns = num_turns0 + num_turns1
    num_particles = len(part.x)

    table01 = record0.table1
    table02 = record0.table2
    num_recorded_tab01 = table01._index.num_recorded
    num_recorded_tab02 = table02._index.num_recorded

    assert num_recorded_tab01 == (num_particles * num_turns)
    assert num_recorded_tab02 == (num_particles * num_turns * (n_kicks0))

    assert np.sum((table01.at_element[:num_recorded_tab01] == 0)) == (num_particles * num_turns)
    assert np.sum((table01.at_element[:num_recorded_tab01] == 1)) == 0
    assert np.sum((table02.at_element[:num_recorded_tab02] == 0)) == (num_particles * num_turns
                                            * n_kicks0)
    assert np.sum((table02.at_element[:num_recorded_tab02] == 1)) == 0

    for i_turn in range(num_turns):
        assert np.sum((table01.at_turn[:num_recorded_tab01] == i_turn)) == num_particles
        assert np.sum((table02.at_turn[:num_recorded_tab02] == i_turn)) == (num_particles
                                                            * (n_kicks0))

    table11 = record1.table1
    table12 = record1.table2
    num_recorded_tab11 = table11._index.num_recorded
    num_recorded_tab12 = table12._index.num_recorded

    assert num_recorded_tab11 == (num_particles * num_turns)
    assert num_recorded_tab12 == (num_particles * num_turns * (n_kicks1))

    assert np.sum((table11.at_element[:num_recorded_tab11] == 0)) == 0
    assert np.sum((table11.at_element[:num_recorded_tab11] == 1)) == (num_particles * num_turns)
    assert np.sum((table12.at_element[:num_recorded_tab12] == 0)) == 0
    assert np.sum((table12.at_element[:num_recorded_tab12] == 1)) == (num_particles * num_turns
                                            * n_kicks1)

    for i_turn in range(num_turns):
        assert np.sum((table11.at_turn[:num_recorded_tab11] == i_turn)) == num_particles
        assert np.sum((table12.at_turn[:num_recorded_tab12] == i_turn)) == (num_particles
                                                            * (n_kicks1))



import numpy as np
import xtrack as xt
import xobjects as xo
from cpymad.madx import Madx

mad = Madx()
mad.input('''
    ang = 0.6;
    lb = 0.5;
    rb: rbend, l=lb, angle=ang;

    beam, particle=proton, energy=1.0;
    seq: sequence, l=2;
    rb, at=1;
    endsequence;
    use, sequence=seq;
''')
tw = xt.Table(mad.twiss(betx=1.0, bety=1.0))

ds_madx = tw['s', 'rb:1'] - tw['s', 'rb:1<<1']

line = xt.Line.from_madx_sequence(mad.sequence.seq)

xo.assert_allclose(ds_madx, line['rb'].length, atol=0, rtol=1e-12)

env = xt.Environment()
env['lb'] = 0.5
env['ang'] = 0.6
env.new('rb_rbarc', xt.Bend)
env.set('rb_rbarc', length='lb', angle='ang', rbend=True, rbarc=True, k0_from_h=True)

xo.assert_allclose(env['rb_rbarc'].length, ds_madx, atol=0, rtol=1e-12)
xo.assert_allclose(env['rb_rbarc'].h * env['rb_rbarc'].length, 0.6, atol=0, rtol=1e-12)
xo.assert_allclose(env['rb_rbarc'].edge_entry_angle, 0.3, atol=0, rtol=1e-12)
xo.assert_allclose(env['rb_rbarc'].edge_exit_angle, 0.3, atol=0, rtol=1e-12)

assert env['rb_rbarc'].get_expr('length') == "(vars['lb'] / f['sinc']((0.5 * vars['ang'])))"
assert env['rb_rbarc'].get_expr('h') == "(f['sin']((0.5 * vars['ang'])) / (0.5 * vars['lb']))"
assert env['rb_rbarc'].get_expr('k0') == env['rb_rbarc'].get_expr('h')
assert env['rb_rbarc'].get_expr('edge_entry_angle') == "(0.0 + (vars['ang'] / 2))"
assert env['rb_rbarc'].get_expr('edge_exit_angle') == "(0.0 + (vars['ang'] / 2))"

env.new('rb_norbarc', xt.Bend)
env.set('rb_norbarc', length='lb', angle='ang', rbend=True, rbarc=False, k0_from_h=True)

xo.assert_allclose(env['rb_norbarc'].length, 0.5, atol=0, rtol=1e-12)
xo.assert_allclose(env['rb_norbarc'].h * env['rb_norbarc'].length, 0.6, atol=0, rtol=1e-12)
xo.assert_allclose(env['rb_norbarc'].k0, env['rb_norbarc'].h, atol=0, rtol=1e-12)
xo.assert_allclose(env['rb_norbarc'].edge_entry_angle, 0.3, atol=0, rtol=1e-12)
xo.assert_allclose(env['rb_norbarc'].edge_exit_angle, 0.3, atol=0, rtol=1e-12)

assert env['rb_norbarc'].get_expr('length') == "vars['lb']"
assert env['rb_norbarc'].get_expr('h') == "(vars['ang'] / vars['lb'])"
assert env['rb_norbarc'].get_expr('k0') == env['rb_norbarc'].get_expr('h')
assert env['rb_norbarc'].get_expr('edge_entry_angle') == "(0.0 + (vars['ang'] / 2))"
assert env['rb_norbarc'].get_expr('edge_exit_angle') == "(0.0 + (vars['ang'] / 2))"

# Check an sbend
env.new('sb', xt.Bend)
env.set('sb', length='lb', angle='ang', rbend=False, rbarc=True, k0_from_h=True)

xo.assert_allclose(env['sb'].length, 0.5, atol=0, rtol=1e-12)
xo.assert_allclose(env['sb'].h * env['sb'].length, 0.6, atol=0, rtol=1e-12)
xo.assert_allclose(env['sb'].edge_entry_angle, 0, atol=1e-12, rtol=0)
xo.assert_allclose(env['sb'].edge_exit_angle, 0, atol=1e-12, rtol=0)

assert env['sb'].get_expr('length') == "vars['lb']"
assert env['sb'].get_expr('h') == "(vars['ang'] / vars['lb'])"
assert env['sb'].get_expr('k0') == env['sb'].get_expr('h')
assert env['sb'].get_expr('edge_entry_angle') == None
assert env['sb'].get_expr('edge_exit_angle') == None


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import xtrack as xt
import xpart as xp
import xobjects as xo

context = xo.ContextCpu()

with open('../../test_data/hllhc15_noerrors_nobb/line_and_particle.json') as f:
    dct = json.load(f)
line = xt.Line.from_dict(dct['line'])
line.particle_ref = xt.Particles.from_dict(dct['particle'])

num_particles = 50
monitor_ip5 = xt.ParticlesMonitor(start_at_turn=5, stop_at_turn=15,
                                    num_particles=num_particles)
monitor_ip8 = xt.ParticlesMonitor(start_at_turn=5, stop_at_turn=15,
                                    num_particles=num_particles)
line.insert_element(index='ip5', element=monitor_ip5, name='mymon5')
line.insert_element(index='ip8', element=monitor_ip8, name='mymon8')

line.build_tracker()

particles = xp.generate_matched_gaussian_bunch(line=line,
                                               num_particles=num_particles,
                                               nemitt_x=2.5e-6,
                                               nemitt_y=2.5e-6,
                                               sigma_z=9e-2)

num_turns = 30
monitor = xt.ParticlesMonitor(_context=context,
                              start_at_turn=5, stop_at_turn=15,
                              num_particles=num_particles)
line.track(particles, num_turns=num_turns)

# monitor_ip5 contains the data recorded in before the element 'ip5', while
# monitor_ip8 contains the data recorded in before the element 'ip8'
# The element index at which the recording is made can be inspected in
# monitor_ip5.at_element.


import math
import numpy as np
from collections import OrderedDict
from pathlib import Path

import pytest
import xobjects as xo
from cpymad.madx import Madx

from xdeps.refs import CompactFormatter
import xtrack as xt
from xtrack import Strategy, Uniform
from xtrack.mad_parser.loader import MadxLoader
from xtrack.mad_parser.parse import MadxOutputType, MadxParser

test_data_folder = (Path(__file__).parent / '../test_data').absolute()


def test_simple_parser():
    sequence = """
    if (version>=50401){option,-rbarc;};  ! to be ignored
    
    third = 1 / 3;
    power = 3^4;
    hello := third * twopi;
    mb.l = 1;
    qd.l = 0.5;
    offset := 0.1;

    mb: sbend, l := mb.l, angle := hello;
    qf: quadrupole, l := 1, k1 := 1;
    qd: quadrupole, l := 1, k1 := -1;
    
    line: sequence, l = 12;
        ip1: marker, at = 0;
        qf1: qf, at := 1 + offset, from = ip1, slot_id = 1;
        mb1: mb, at := 2 + offset, from = ip1, slot_id = 2;
        qd1: qd, at := 3 + offset, from = ip1, slot_id = 3;
    endsequence;

    mb1, k0 := hello, polarity = +1;
    qf1, knl := {0, 0, 0, 0.01, 0};
    qd1, knl := {0, 0, 0, -0.01, 0};
    
    return;  ! should also be ignored
    """

    env = xt.Environment()
    env['twopi'] = 2 * math.pi
    parser = MadxParser(vars=env.vars, functions=env.functions)
    result = parser.parse_string(sequence)

    expected = {
        'elements': {
            'mb': {
                'angle': env.vars['hello'],
                'l': env.vars['mb.l'],
                'parent': 'sbend',
            },
            'qf': {
                'k1': 1.0,
                'l': 1.0,
                'parent': 'quadrupole',
            },
            'qd': {
                'k1': -1.0,
                'l': 1.0,
                'parent': 'quadrupole',
            },
        },
        'lines': {
            'line': {
                'l': 12.0,
                'parent': 'sequence',
                'elements': {
                    'ip1': {
                        'at': 0.0,
                        'parent': 'marker',
                    },
                    'qf1': {
                        'at': 1.0 + env.vars['offset'],
                        'from': 'ip1',
                        'parent': 'qf',
                        'slot_id': 1.0,
                    },
                    'mb1': {
                        'at': 2.0 + env.vars['offset'],
                        'from': 'ip1',
                        'parent': 'mb',
                        'slot_id': 2.0,
                    },
                    'qd1': {
                        'at': 3.0 + env.vars['offset'],
                        'from': 'ip1',
                        'parent': 'qd',
                        'slot_id': 3.0,
                    },
                },
            },
        },
        'parameters': {
            'mb1': {
                'k0': env.vars['hello'],
                'polarity': 1.0,
            },
            'qf1': {
                'knl': [0.0, 0.0, 0.0, 0.01, 0.0],
            },
            'qd1': {
                'knl': [0.0, 0.0, 0.0, -0.01, 0.0],
            },
        },
    }

    def _order_madx_output(item):
        item['elements'] = OrderedDict(item['elements'])
        item['lines'] = OrderedDict(item['lines'])
        for line in item['lines'].values():
            line['elements'] = OrderedDict(line['elements'])
        item['parameters'] = OrderedDict(item['parameters'])

    _order_madx_output(expected)
    _order_madx_output(result)

    assert expected == result


@pytest.mark.parametrize(
    'input,value,expr',
    [
        ('a := 42;', 42, None),
        ('c := 3; d := 4; a := c^d;', 81, '(c ** d)'),
        ('c := 3; d := 4; a = c^d;', 81, None),
    ]
)
def test_parse_simple_expression(input, value, expr):
    env = xt.load_madx_lattice(string=input)
    assert env['a'] == value

    if expr is not None:
        formatter = CompactFormatter(None)
        assert env.get_expr('a')._formatted(formatter) == expr
    else:
        assert env.get_expr('a') is None
        env['c'] = 8
        assert env['a'] == value


@pytest.fixture(scope='module')
def example_sequence(temp_context_default_mod):
    sequence = """
    ll = 36;

    vk: vkicker, l=2, kick=3, tilt=-1;
    hk: hkicker, l=1, kick=6, tilt=-2;
    ki: kicker, l=2, vkick=3, hkick=4, tilt=1;
    tk: tkicker, l=1, vkick=3, hkick=4, tilt=2;
    in: instrument, l=2;
    mo: monitor, l=1;
    pl: placeholder, l=1;
    sb: sbend, l=2, angle=2, tilt=-2, k0=3, k1=1, k2=2, k1s=3, e1=2, e2=1, fint=3, fintx=2, hgap=1, h1=3, h2=2;
    rb: rbend, l=1.5, angle=2, tilt=-2, k0=3, k1=1, k2=2, k1s=3, e1=2, e2=1, fint=3, fintx=2, hgap=1, h1=3, h2=2;
    qu: quadrupole, l=2, k1=3, k1s=4, tilt=2;  ! ignore thick and ktap
    se: sextupole, L=1, K2=2, K2S=3, TILT=2;  ! ignore ktap
    oc: octupole, L=2, K3=3, K3S=2, TILT=2;
    ma: marker;
    rf: rfcavity, L=2, VOLT=1, LAG=2, FREQ=3, HARMON=2, NO_CAVITY_TOTALPATH;  ! ignore N_BESSEL
    mu: multipole, LRAD=1, TILT=2, KNL={3, 4, 5, 6}, KSL={1, 2, 3, 4};
    so: solenoid, l=2, ks=3;  ! ignore ksi
    
    rb_stage1: rbend, l=1;
    rb_stage2: rb_stage1, angle=2;

    ! Not yet implemented
    ! co: collimator, l=2, apertype=ellipse, aperture={0.01,0.005}, aper_offset={x,y}, aper_tol={corner_r,inner_x,inner_y};

    line: sequence, l = ll;
        vk1: vk, at = 1;
        hk1: hk, at = 3;
        ki1: ki, at = 5;
        tk1: tk, at = 7;
        in1: in, at = 9;
        mo1: mo, at = 11;
        pl1: pl, at = 13;
        sb1: sb, at = 15;
        rb1: rb, at = 17;
        qu1: qu, at = 19;
        se1: se, at = 21;
        oc1: oc, at = 23;
        ma1: ma, at = 25;
        rf1: rf, at = 27;
        mu1: mu, at = 29;
        so1: so, at = 31;

        rx1: rb_stage2, at = 33;
        rx2: rb_stage2, at = 35;
    endsequence;
    
    ! exactly the same as above, but to be parsed in reverse order
    line_reversed: sequence, l = ll;
        vk1: vk, at = 1;
        hk1: hk, at = 3;
        ki1: ki, at = 5;
        tk1: tk, at = 7;
        in1: in, at = 9;
        mo1: mo, at = 11;
        pl1: pl, at = 13;
        sb1: sb, at = 15;
        rb1: rb, at = 17;
        qu1: qu, at = 19;
        se1: se, at = 21;
        oc1: oc, at = 23;
        ma1: ma, at = 25;
        rf1: rf, at = 27;
        mu1: mu, at = 29;
        so1: so, at = 31;

        rx1: rb_stage2, at = 33;

    endsequence;

    rx2, angle = 1.5;
    """

    env = xt.load_madx_lattice(string=sequence, reverse_lines=['line_reversed'])

    def make_positions(line):
        tt = line.get_table()
        diffs = np.diff(tt['s'], append=tt['s', '_end_point'])
        s_centre = tt['s'] + 0.5 * diffs
        return dict(zip(tt['name'], s_centre))

    positions = make_positions(env['line'])
    positions_reversed = make_positions(env['line_reversed'])

    return env, positions, positions_reversed


def test_parsed_lines(example_sequence):
    env, _, _ = example_sequence
    assert env['line'].name == 'line'
    assert env['line_reversed'].name == 'line_reversed'


def test_vkick(example_sequence):
    env, positions, _ = example_sequence
    vk1 = env['vk1/line']
    xo.assert_allclose(positions['vk1/line'], 1)
    assert isinstance(vk1, xt.Multipole)
    assert vk1.length == 2
    assert vk1.knl[0] == 0
    assert vk1.ksl[0] == 3
    xo.assert_allclose(vk1.rot_s_rad, -1)


def test_hkick(example_sequence):
    env, positions, _ = example_sequence
    hk1 = env['hk1/line']
    xo.assert_allclose(positions['hk1/line'], 3)
    assert isinstance(hk1, xt.Multipole)
    assert hk1.length == 1
    assert hk1.knl[0] == -6
    assert hk1.ksl[0] == 0
    assert hk1.rot_s_rad == -2


def test_kick(example_sequence):
    env, positions, _ = example_sequence
    ki1 = env['ki1/line']
    xo.assert_allclose(positions['ki1/line'], 5)
    assert isinstance(ki1, xt.Multipole)
    assert ki1.length == 2
    assert ki1.knl[0] == -4
    assert ki1.ksl[0] == 3
    xo.assert_allclose(ki1.rot_s_rad, 1)


def test_tkick(example_sequence):
    env, positions, _ = example_sequence
    tk1 = env['tk1/line']
    xo.assert_allclose(positions['tk1/line'], 7)
    assert isinstance(tk1, xt.Multipole)
    assert tk1.length == 1
    assert tk1.knl[0] == -4
    assert tk1.ksl[0] == 3
    assert tk1.rot_s_rad == 2


def test_instrument(example_sequence):
    env, positions, _ = example_sequence
    in1 = env['in1/line']
    xo.assert_allclose(positions['in1/line'], 9)
    assert isinstance(in1, xt.Drift)
    assert in1.length == 2


def test_monitor(example_sequence):
    env, positions, _ = example_sequence
    mo1 = env['mo1/line']
    xo.assert_allclose(positions['mo1/line'], 11)
    assert isinstance(mo1, xt.Drift)
    assert mo1.length == 1


def test_placeholder(example_sequence):
    env, positions, _ = example_sequence
    pl1 = env['pl1/line']
    xo.assert_allclose(positions['pl1/line'], 13)
    assert isinstance(pl1, xt.Drift)
    assert pl1.length == 1


def test_sbend(example_sequence):
    env, positions, _ = example_sequence
    # sb: sbend, l=2, angle=2, tilt=-2, k0=3, k1=1, k2=2, k1s=3, e1=2, e2=1,
    #   fint=3, fintx=2, hgap=1;  ! thick, ktap, h1, h2 we ignore
    sb1 = env['sb1/line']
    xo.assert_allclose(positions['sb1/line'], 15)
    assert isinstance(sb1, xt.Bend)
    assert sb1.length == 2
    assert sb1.k0 == 3
    assert sb1.h == 2 / 2  # angle / l
    assert sb1.k1 == 1
    assert sb1.knl[0] == 0
    assert sb1.knl[1] == 0
    assert sb1.knl[2] == 2 * 2  # k2 * l
    assert sb1.ksl[0] == 0
    assert sb1.ksl[1] == 3 * 2  # k1s * l
    assert sb1.edge_entry_angle == 2
    assert sb1.edge_exit_angle == 1
    assert sb1.edge_entry_fint == 3
    assert sb1.edge_exit_fint == 2
    assert sb1.edge_entry_hgap == 1
    assert sb1.edge_exit_hgap == 1


def test_rbend(example_sequence):
    env, positions, _ = example_sequence
    # rb: rbend, l=2, angle=1.5, tilt=-2, k0=3, k1=1, k2=2, k1s=3, e1=2, e2=1,
    #   fint=3, fintx=2, hgap=1, h1=3, h2=2;  ! ditto
    rb1 = env['rb1/line']
    xo.assert_allclose(positions['rb1/line'], 17)
    assert isinstance(rb1, xt.RBend)

    angle = 2
    l = 1.5
    R = 0.5 * l / math.sin(0.5 * angle)
    l_curv = R * angle
    h = 1 / R

    assert rb1.length == l_curv
    assert rb1.k0 == 3
    assert rb1.h == h
    assert rb1.k1 == 1
    assert rb1.knl[0] == 0
    assert rb1.knl[1] == 0
    assert rb1.knl[2] == 2 * l  # k2 * l
    assert rb1.ksl[0] == 0
    assert rb1.ksl[1] == 3 * l  # k1s * l
    assert rb1.edge_entry_angle == 2
    assert rb1.edge_exit_angle == 1
    assert rb1.edge_entry_fint == 3
    assert rb1.edge_exit_fint == 2
    assert rb1.edge_entry_hgap == 1
    assert rb1.edge_exit_hgap == 1

def test_rbend_two_step(example_sequence):
    env, positions, _ = example_sequence
    rb1 = env['rx1/line']
    xo.assert_allclose(positions['rx1/line'], 33)
    assert isinstance(rb1, xt.RBend)

    angle = 2
    l = 1
    R = 0.5 * l / math.sin(0.5 * angle)
    l_curv = R * angle
    h = 1 / R

    assert rb1.length == l_curv
    assert rb1.h == h
    assert rb1.edge_entry_angle == 0
    assert rb1.edge_exit_angle == 0
    assert rb1.k0 == h

def test_rbend_set_params_after_lattice(example_sequence):
    env, positions, _ = example_sequence
    rb1 = env['rx2']
    assert positions['rx2'] < 35 # When changing the angle the position changes
                                 # (the magnet shortens)
    assert isinstance(rb1, xt.RBend)

    angle = 1.5
    l = 1
    R = 0.5 * l / math.sin(0.5 * angle)
    l_curv = R * angle
    h = 1 / R

    assert rb1.length == l_curv
    assert rb1.h == h
    assert rb1.edge_entry_angle == 0
    assert rb1.edge_exit_angle == 0
    assert rb1.k0 == h


def test_quadrupole(example_sequence):
    env, positions, _ = example_sequence
    # qu: quadrupole, l=2, k1=3, k1s=4, tilt=2;  ! ignore thick and ktap
    qu1 = env['qu1/line']
    xo.assert_allclose(positions['qu1/line'], 19)
    assert isinstance(qu1, xt.Quadrupole)
    assert qu1.length == 2
    assert qu1.k1 == 3
    assert qu1.k1s == 4
    assert qu1.rot_s_rad == 2


def test_sextupole(example_sequence):
    env, positions, _ = example_sequence
    # se: sextupole, L=1, K2=2, K2S=3, TILT=2;  ! ignore ktap
    se1 = env['se1/line']
    xo.assert_allclose(positions['se1/line'], 21)
    assert isinstance(se1, xt.Sextupole)
    assert se1.length == 1
    assert se1.k2 == 2
    assert se1.k2s == 3
    assert se1.rot_s_rad == 2


def test_octupole(example_sequence):
    env, positions, _ = example_sequence
    # oc: octupole, L=2, K3=3, K3S=2, TILT=2;
    oc1 = env['oc1/line']
    xo.assert_allclose(positions['oc1/line'], 23)
    assert isinstance(oc1, xt.Octupole)
    assert oc1.length == 2
    assert oc1.k3 == 3
    assert oc1.k3s == 2
    assert oc1.rot_s_rad == 2


def test_marker(example_sequence):
    env, positions, _ = example_sequence
    # ma: marker;
    ma1 = env['ma1']
    xo.assert_allclose(positions['ma1'], 25)
    assert isinstance(ma1, xt.Marker)


def test_rfcavity(example_sequence):
    env, positions, _ = example_sequence
    # rf: rfcavity, L=2, VOLT=1, LAG=2, FREQ=3, HARMON=2;  ! ignore N_BESSEL, NO_CAVITY_TOTALPATH
    rf1 = env['rf1/line']
    xo.assert_allclose(positions['rf1/line'], 27)
    assert isinstance(rf1, xt.Cavity)
    assert rf1.voltage == 1e6
    assert rf1.lag == 2 * 360
    assert rf1.frequency == 3e6


def test_multipole(example_sequence):
    env, positions, _ = example_sequence
    # mu: multipole, LRAD=1, TILT=2, KNL={3, 4, 5}, KSL={1, 2, 3};
    mu1 = env['mu1/line']
    xo.assert_allclose(positions['mu1/line'], 29)
    assert isinstance(mu1, xt.Multipole)
    assert mu1.length == 1
    assert mu1.knl[0] == 3
    assert mu1.knl[1] == 4
    assert mu1.knl[2] == 5
    assert mu1.knl[3] == 6
    assert mu1.ksl[0] == 1
    assert mu1.ksl[1] == 2
    assert mu1.ksl[2] == 3
    assert mu1.ksl[3] == 4
    assert mu1.rot_s_rad == 2


def test_solenoid(example_sequence):
    env, positions, _ = example_sequence
    # so: solenoid, l=2, ks=3;  ! ignore ksi
    so1 = env['so1/line']
    xo.assert_allclose(positions['so1/line'], 31)
    assert isinstance(so1, xt.Solenoid)
    assert so1.length == 2
    assert so1.ks == 3


def test_reversed_vkick(example_sequence):
    env, _, positions_reversed = example_sequence
    ivk1 = env['vk1/line_reversed']
    xo.assert_allclose(positions_reversed['vk1/line_reversed'], 36 - 1)
    assert isinstance(ivk1, xt.Multipole)
    assert ivk1.length == 2
    assert ivk1.knl[0] == 0
    assert ivk1.ksl[0] == -3
    xo.assert_allclose(ivk1.rot_s_rad,  1)


def test_reversed_hkick(example_sequence):
    env, _, positions_reversed = example_sequence
    hk1 = env['hk1/line_reversed']
    xo.assert_allclose(positions_reversed['hk1/line_reversed'], 36 - 3)
    assert isinstance(hk1, xt.Multipole)
    assert hk1.length == 1
    assert hk1.knl[0] == -6
    assert hk1.ksl[0] == 0
    assert hk1.rot_s_rad == 2


def test_reversed_kick(example_sequence):
    env, _, positions_reversed = example_sequence
    ki1 = env['ki1/line_reversed']
    xo.assert_allclose(positions_reversed['ki1/line_reversed'], 36 - 5)
    assert isinstance(ki1, xt.Multipole)
    assert ki1.length == 2
    assert ki1.knl[0] == -4
    assert ki1.ksl[0] == -3
    xo.assert_allclose(ki1.rot_s_rad, -1)


def test_reversed_tkick(example_sequence):
    env, _, positions_reversed = example_sequence
    tk1 = env['tk1/line_reversed']
    xo.assert_allclose(positions_reversed['tk1/line_reversed'], 36 - 7)
    assert isinstance(tk1, xt.Multipole)
    assert tk1.length == 1
    assert tk1.knl[0] == -4
    assert tk1.ksl[0] == -3
    assert tk1.rot_s_rad == -2


def test_reversed_instrument(example_sequence):
    env, _, positions_reversed = example_sequence
    in1 = env['in1/line_reversed']
    xo.assert_allclose(positions_reversed['in1/line_reversed'], 36 - 9)
    assert isinstance(in1, xt.Drift)
    assert in1.length == 2


def test_reversed_monitor(example_sequence):
    env, _, positions_reversed = example_sequence
    mo1 = env['mo1/line_reversed']
    xo.assert_allclose(positions_reversed['mo1/line_reversed'], 36 - 11)
    assert isinstance(mo1, xt.Drift)
    assert mo1.length == 1


def test_reversed_placeholder(example_sequence):
    env, _, positions_reversed = example_sequence
    pl1 = env['pl1/line_reversed']
    xo.assert_allclose(positions_reversed['pl1/line_reversed'], 36 - 13)
    assert isinstance(pl1, xt.Drift)
    assert pl1.length == 1


def test_reversed_sbend(example_sequence):
    env, _, positions_reversed = example_sequence
    # sb: sbend, l=2, angle=2, tilt=-2, k0=3, k1=1, k2=2, k1s=3, e1=2, e2=1,
    #   fint=3, fintx=2, hgap=1;  ! thick, ktap, h1, h2 we ignore
    sb1 = env['sb1/line_reversed']
    xo.assert_allclose(positions_reversed['sb1/line_reversed'], 36 - 15)
    assert isinstance(sb1, xt.Bend)
    assert sb1.length == 2
    assert sb1.k0 == 3
    assert sb1.h == 2 / 2  # angle / l
    assert sb1.k1 == -1
    assert sb1.knl[0] == 0
    assert sb1.knl[1] == 0
    assert sb1.knl[2] == 2 * 2  # k2 * l
    assert sb1.ksl[0] == 0
    assert sb1.ksl[1] == 3 * 2  # k1s * l
    assert sb1.edge_entry_angle == 1
    assert sb1.edge_exit_angle == 2
    assert sb1.edge_entry_fint == 2
    assert sb1.edge_exit_fint == 3
    assert sb1.edge_entry_hgap == 1
    assert sb1.edge_exit_hgap == 1


def test_reversed_rbend(example_sequence):
    env, _, positions_reversed = example_sequence    # rb: rbend, l=2, angle=1.5, tilt=-2, k0=3, k1=1, k2=2, k1s=3, e1=2, e2=1,
    #   fint=3, fintx=2, hgap=1, h1=3, h2=2;  ! ditto
    rb1 = env['rb1/line_reversed']
    xo.assert_allclose(positions_reversed['rb1/line_reversed'], 36 - 17)
    assert isinstance(rb1, xt.RBend)

    angle = 2
    l = 1.5
    R = 0.5 * l / math.sin(0.5 * angle)
    l_curv = R * angle
    h = 1 / R

    assert rb1.length == l_curv
    assert rb1.k0 == 3
    assert rb1.h == h
    assert rb1.k1 == -1
    assert rb1.knl[0] == 0
    assert rb1.knl[1] == 0
    assert rb1.knl[2] == 2 * l  # k2 * l
    assert rb1.ksl[0] == 0
    assert rb1.ksl[1] == 3 * l  # k1s * l
    assert rb1.edge_entry_angle == 1
    assert rb1.edge_exit_angle == 2
    assert rb1.edge_entry_fint == 2
    assert rb1.edge_exit_fint == 3
    assert rb1.edge_entry_hgap == 1
    assert rb1.edge_exit_hgap == 1


def test_reversed_quadrupole(example_sequence):
    env, _, positions_reversed = example_sequence
    # qu: quadrupole, l=2, k1=3, k1s=4, tilt=2;  ! ignore thick and ktap
    qu1 = env['qu1/line_reversed']
    xo.assert_allclose(positions_reversed['qu1/line_reversed'], 36 - 19)
    assert isinstance(qu1, xt.Quadrupole)
    assert qu1.length == 2
    assert qu1.k1 == -3
    assert qu1.k1s == 4
    assert qu1.rot_s_rad == -2


def test_reversed_sextupole(example_sequence):
    env, _, positions_reversed = example_sequence
    # se: sextupole, L=1, K2=2, K2S=3, TILT=2;  ! ignore ktap
    se1 = env['se1/line_reversed']
    xo.assert_allclose(positions_reversed['se1/line_reversed'], 36 - 21)
    assert isinstance(se1, xt.Sextupole)
    assert se1.length == 1
    assert se1.k2 == 2
    assert se1.k2s == -3
    assert se1.rot_s_rad == -2


def test_reversed_octupole(example_sequence):
    env, _, positions_reversed = example_sequence
    # oc: octupole, L=2, K3=3, K3S=2, TILT=2;
    oc1 = env['oc1/line_reversed']
    xo.assert_allclose(positions_reversed['oc1/line_reversed'], 36 - 23)
    assert isinstance(oc1, xt.Octupole)
    assert oc1.length == 2
    assert oc1.k3 == -3
    assert oc1.k3s == 2
    assert oc1.rot_s_rad == -2


def test_reversed_marker(example_sequence):
    env, _, positions_reversed = example_sequence    # ma: marker;
    ma1 = env['ma1']
    xo.assert_allclose(positions_reversed['ma1'], 36 - 25)
    assert isinstance(ma1, xt.Marker)


def test_reversed_rfcavity(example_sequence):
    env, _, positions_reversed = example_sequence    # ma: marker;    # rf: rfcavity, L=2, VOLT=1, LAG=2, FREQ=3, HARMON=2;  ! ignore N_BESSEL, NO_CAVITY_TOTALPATH
    rf1 = env['rf1/line_reversed']
    xo.assert_allclose(positions_reversed['rf1/line_reversed'], 36 - 27)
    assert isinstance(rf1, xt.Cavity)
    assert rf1.voltage == 1e6
    assert rf1.lag == 180 - 2 * 360
    assert rf1.frequency == 3e6


def test_reversed_multipole(example_sequence):
    env, _, positions_reversed = example_sequence
    # mu: multipole, LRAD=1, TILT=2, KNL={3, 4, 5}, KSL={1, 2, 3};
    mu1 = env['mu1/line_reversed']
    xo.assert_allclose(positions_reversed['mu1/line_reversed'], 36 - 29)
    assert isinstance(mu1, xt.Multipole)
    assert mu1.length == 1
    assert mu1.knl[0] == 3
    assert mu1.knl[1] == -4
    assert mu1.knl[2] == 5
    assert mu1.knl[3] == -6
    assert mu1.ksl[0] == -1
    assert mu1.ksl[1] == 2
    assert mu1.ksl[2] == -3
    assert mu1.ksl[3] == 4
    assert mu1.rot_s_rad == -2


def test_reversed_solenoid(example_sequence):
    env, _, positions_reversed = example_sequence
    # so: solenoid, l=2, ks=3;  ! ignore ksi
    so1 = env['so1/line_reversed']
    xo.assert_allclose(positions_reversed['so1/line_reversed'], 36 - 31)
    assert isinstance(so1, xt.Solenoid)
    assert so1.length == 2
    assert so1.ks == -3


def test_load_b2_with_bv_minus_one(tmp_path):
    test_data_folder_str = str(test_data_folder)

    mad = Madx(stdout=False)
    mad.call(test_data_folder_str + '/hllhc15_thick/lhc.seq')
    mad.call(test_data_folder_str + '/hllhc15_thick/hllhc_sequence.madx')
    mad.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
    mad.use('lhcb1')
    mad.input('beam, sequence=lhcb2, particle=proton, energy=7000, bv=-1;')
    mad.use('lhcb2')
    mad.call(test_data_folder_str + '/hllhc15_thick/opt_round_150_1500.madx')
    mad.twiss()

    mad.globals['vrf400'] = 16  # Check voltage expressions
    mad.globals['lagrf400.b2'] = 0.02  # Check lag expressions
    mad.globals['on_x1'] = 100  # Check kicker expressions
    mad.globals['on_sep2'] = 2  # Check kicker expressions
    mad.globals['on_x5'] = 123  # Check kicker expressions
    mad.globals['kqtf.b2'] = 1e-5  # Check quad expressions
    mad.globals['ksf.b2'] = 1e-3  # Check sext expressions
    mad.globals['kqs.l3b2'] = 1e-4  # Check skew expressions
    mad.globals['kss.a45b2'] = 1e-4  # Check skew sext expressions
    mad.globals['kof.a34b2'] = 3  # Check oct expressions
    mad.globals['on_crab1'] = -190  # Check cavity expressions
    mad.globals['on_crab5'] = -130  # Check cavity expressions
    mad.globals['on_sol_atlas'] = 1  # Check solenoid expressions
    mad.globals['kcdx3.r1'] = 1e-4  # Check thin decapole expressions
    mad.globals['kcdsx3.r1'] = 1e-4  # Check thin skew decapole expressions
    mad.globals['kctx3.l1'] = 1e-5  # Check thin dodecapole expressions
    mad.globals['kctsx3.r1'] = 1e-5  # Check thin skew dodecapole expressions


    tmp_seq_path = str(tmp_path / 'sequence.seq')
    mad.input('set, format=".20g";')
    mad.save(file=tmp_seq_path)

    line2_ref = xt.Line.from_madx_sequence(
        sequence=mad.sequence.lhcb2,
        allow_thick=True,
        deferred_expressions=True,
        replace_in_expr={'bv_aux': 'bvaux_b2'},
    )
    line2_ref.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7000e9)

    env = xt.load_madx_lattice(tmp_seq_path, reverse_lines=['lhcb2'])
    line2 = env['lhcb2']

    # Remove apertures, they are not supported in the cpymadloader
    line2.element_names = [
        name for name in line2.element_names
        if not isinstance(line2[name], xt.LimitRectEllipse)
    ]

    # Bend done

    # Quadrupole
    xo.assert_allclose(line2_ref['mq.27l2.b2'].k1, line2['mq.27l2.b2'].k1, rtol=0, atol=1e-12)
    xo.assert_allclose(line2_ref['mqs.27l3.b2'].k1s, line2['mqs.27l3.b2'].k1s, rtol=0, atol=1e-12)

    tt2 = line2_ref.get_table()
    tt4 = line2.get_table()

    tt2nodr = tt2.rows[tt2.element_type != 'Drift']
    tt4nodr = tt4.rows[tt4.element_type != 'Drift']

    # Check s
    l2names = list(tt2nodr.name)
    l4names = list(tt4nodr.name)

    l2names.remove('lhcb2$start')
    l2names.remove('lhcb2$end')

    assert l2names == [nn[:-len('/lhcb2')] if nn.endswith('/lhcb2') else nn for nn in l4names]

    # The tolerance is higher than usual, due to the difference in handling
    # rbend lengths between MAD-X and Xtrack (RBends' straight length is
    # the invariant in Xtrack when changing the angle, which is not the case
    # for MAD-X with `rbarc` == False).
    xo.assert_allclose(tt2nodr.rows[l2names].s, tt4nodr.rows[l4names].s, rtol=0, atol=1.5e-6)

    for nn in l4names:
        if nn == '_end_point':
            continue
        nn_straight = nn[:-len('/lhcb2')] if nn.endswith('/lhcb2') else nn
        e2 = line2_ref[nn_straight]
        e4 = line2[nn]
        d2 = e2.to_dict()
        d4 = e4.to_dict()
        is_rbend = isinstance(e4, xt.RBend)

        for kk in d2.keys():
            if kk in ('__class__', 'model', 'side'):
                assert d2[kk] == d4[kk]
                continue

            if kk in {
                'order',  # Always assumed to be 5, not always the same
                'frequency',  # If not specified, depends on the beam,
                              # so for now we ignore it
            }:
                continue

            if kk in {'knl', 'ksl'}:
                maxlen = max(len(d2[kk]), len(d4[kk]))
                lhs = np.pad(d2[kk], (0, maxlen - len(d2[kk])), mode='constant')
                rhs = np.pad(d4[kk], (0, maxlen - len(d4[kk])), mode='constant')
                xo.assert_allclose(lhs, rhs, rtol=1e-10, atol=1e-16)
                continue

            if is_rbend and kk in ('length', 'length_straight'):
                xo.assert_allclose(d2[kk], d4[kk], rtol=1e-7, atol=1e-6)
                continue

            if is_rbend and kk in ('h', 'k0'):
                xo.assert_allclose(d2[kk], d4[kk], rtol=1e-7, atol=5e-10)
                continue

            xo.assert_allclose(d2[kk], d4[kk], rtol=1e-10, atol=1e-16)


def test_line_syntax():
    sequence = """
    el1: drift, l=1;
    el2: drift, l=2;
    el3: drift, l=3;
    
    l1: line = (el1, el2, el3);
    l2: line = (-l1);
    l3: line = (3 * el1, 2 * el2);
    l4: line = (-l3, l3, el3);
    l5: line = (-2 * l4);
    l6: line = (3 * (el1, el2), -(el2, el1));
    """

    loader = MadxLoader()
    loader.load_string(sequence)
    env = loader.env

    l1 = env['l1']
    assert l1.name == 'l1'
    assert l1.element_names == ['el1', 'el2', 'el3']

    l2 = env['l2']
    assert l2.name == 'l2'
    assert l2.element_names == ['el3', 'el2', 'el1']

    l3 = env['l3']
    assert l3.name == 'l3'
    assert l3.element_names == 3 * ['el1'] + 2 * ['el2']

    l4 = env['l4']
    assert l4.name == 'l4'
    assert l4.element_names == 2 * ['el2'] + 6 * ['el1'] + 2 * ['el2'] + ['el3']

    l5 = env['l5']
    assert l5.name == 'l5'
    assert l5.element_names == 2 * (['el3'] + 2 * ['el2'] + 6 * ['el1'] + 2 * ['el2'])

    l6 = env['l6']
    assert l6.name == 'l6'
    assert l6.element_names == 4 * ['el1', 'el2']


def test_refer_and_thin_elements():
    sequence = """
    mb: sbend, l = 3;
    cav: rfcavity, l = 4;
    seq1: sequence, l = 10;
        mb1: mb, at = 1.5;
        cav1: cav, at = 5;
        mb2: mb, at = 8.5;
    endsequence;

    seq2: sequence, l = 10, refer = ENTRY;
        mb1: mb, at = 0;
        cav1: cav, at = 3;
        mb2: mb, at = 7;
    endsequence;
    """

    loader = MadxLoader()
    loader.load_string(sequence)
    env = loader.env

    seq1 = env['seq1']
    seq1.merge_consecutive_drifts()
    tt1 = seq1.get_table()

    seq2 = env['seq2']
    seq2.merge_consecutive_drifts()
    tt2 = seq2.get_table()

    assert np.all(tt1['element_type'] == tt2['element_type'])
    assert np.all(tt1['s'] == tt2['s'])


def test_import_seq_length():
    sequence = """
    qu: quadrupole, l=2, k1=3, k1s=4, tilt=2;  ! ignore thick and ktap

    ll := 3 * a;
    a = 10;

    line: sequence, l = ll;
        qu1: qu, at = 19;
    endsequence;
    """

    env = xt.load_madx_lattice(string=sequence)

    tt = env.line.get_table()
    assert np.all(tt.name == np.array(['drift_1', 'qu1', 'drift_2', '_end_point']))
    xo.assert_allclose(tt['s'], np.array([ 0., 18., 20., 30.]), rtol=0, atol=1e-15)
    assert env.line.builder.length == 30


def test_repeated_element_mad_behaviour():
    sequence = """
    mar: marker;
    ben: sbend, l=1;
    
    seq1: sequence, l=10;
      ee: mar, at=5;
    endsequence;
    
    seq2: sequence, l=10;
      ee: ben, at=5;  ! in MAD-X this definition will be ignored
    endsequence;
    """

    env = xt.load_madx_lattice(string=sequence)

    element = env['ee']
    assert env.seq1['ee'] == element
    assert env.seq2['ee'] == element


@pytest.mark.parametrize('aper_config', ['attached_to_marker', 'standalone'])
def test_apertures_on_markers(aper_config):
    if aper_config == 'attached_to_marker':
        sequence = """
            ! Attached to a marker
            m_circle: marker, apertype="circle", aperture={.2};
            m_ellipse: marker, apertype="ellipse", aperture={.2, .1};
            m_rectangle: marker, apertype="rectangle", aperture={.07, .05};
            m_rectellipse: marker, apertype="rectellipse", aperture={.2, .4, .25, .45};
            m_racetrack: marker, apertype="racetrack", aperture={.6,.4,.2,.1};
            m_octagon: marker, apertype="octagon", aperture={.4, .5, 0.5, 1.};
            m_polygon: marker, apertype="circle", aper_vx={+5.800e-2,+5.800e-2,-8.800e-2}, aper_vy={+3.500e-2,-3.500e-2,+0.000e+0};
            """
    else:
        sequence = """
            ! Standalone
            m_circle: circle, aperture={.2};
            m_ellipse: ellipse, aperture={.2, .1};
            m_rectangle: rectangle, aperture={.07, .05};
            m_rectellipse: rectellipse, aperture={.2, .4, .25, .45};
            m_racetrack: racetrack, aperture={.6,.4,.2,.1};
            m_octagon: octagon, aperture={.4, .5, 0.5, 1.};
            m_polygon: circle, aper_vx={+5.800e-2,+5.800e-2,-8.800e-2}, aper_vy={+3.500e-2,-3.500e-2,+0.000e+0};
            """

    sequence += """
        line: sequence, l=1;
            m_circle, at=0;
            m_ellipse, at=0.01;
            m_rectangle, at=0.02;
            m_rectellipse, at=0.03;
            m_racetrack, at=0.04;
            m_octagon, at=0.05;
            m_polygon, at=0.06;
        endsequence;
        """

    env = xt.load_madx_lattice(string=sequence)
    line = env.line

    apertures = [ee for ee in line.elements if ee.__class__.__name__.startswith('Limit')]

    assert isinstance(line[line['m_circle'].name_associated_aperture], xt.LimitEllipse)
    circ = apertures[0]
    assert circ.__class__.__name__ == 'LimitEllipse'
    xo.assert_allclose(circ.a_squ, .2**2, atol=1e-13, rtol=0)
    xo.assert_allclose(circ.b_squ, .2**2, atol=1e-13, rtol=0)

    assert isinstance(line[line['m_ellipse'].name_associated_aperture], xt.LimitEllipse)
    ellip = apertures[1]
    assert ellip.__class__.__name__ == 'LimitEllipse'
    xo.assert_allclose(ellip.a_squ, .2**2, atol=1e-13, rtol=0)
    xo.assert_allclose(ellip.b_squ, .1**2, atol=1e-13, rtol=0)

    assert isinstance(line[line['m_rectangle'].name_associated_aperture], xt.LimitRect)
    rect = apertures[2]
    assert rect.__class__.__name__ == 'LimitRect'
    assert rect.min_x == -.07
    assert rect.max_x == +.07
    assert rect.min_y == -.05
    assert rect.max_y == +.05

    assert isinstance(line[line['m_rectellipse'].name_associated_aperture], xt.LimitRectEllipse)
    rectellip = apertures[3]
    assert rectellip.max_x == .2
    assert rectellip.max_y == .4
    xo.assert_allclose(rectellip.a_squ, .25**2, atol=1e-13, rtol=0)
    xo.assert_allclose(rectellip.b_squ, .45**2, atol=1e-13, rtol=0)

    assert isinstance(line[line['m_racetrack'].name_associated_aperture], xt.LimitRacetrack)
    racetr = apertures[4]
    assert racetr.__class__.__name__ == 'LimitRacetrack'
    assert racetr.min_x == -.6
    assert racetr.max_x == +.6
    assert racetr.min_y == -.4
    assert racetr.max_y == +.4
    assert racetr.a == .2
    assert racetr.b == .1

    assert isinstance(line[line['m_octagon'].name_associated_aperture], xt.LimitPolygon)
    octag = apertures[5]
    assert octag.__class__.__name__ == 'LimitPolygon'
    assert octag._xobject.x_vertices[0] == 0.4
    xo.assert_allclose(octag._xobject.y_vertices[0], 0.4*np.tan(0.5), atol=1e-14, rtol=0)
    assert octag._xobject.y_vertices[1] == 0.5
    xo.assert_allclose(octag._xobject.x_vertices[1], 0.5/np.tan(1.), atol=1e-14, rtol=0)

    assert octag._xobject.y_vertices[2] == 0.5
    xo.assert_allclose(octag._xobject.x_vertices[2], -0.5/np.tan(1.), atol=1e-14, rtol=0)
    assert octag._xobject.x_vertices[3] == -0.4
    xo.assert_allclose(octag._xobject.y_vertices[3], 0.4*np.tan(0.5), atol=1e-14, rtol=0)


    assert octag._xobject.x_vertices[4] == -0.4
    xo.assert_allclose(octag._xobject.y_vertices[4], -0.4*np.tan(0.5), atol=1e-14, rtol=0)
    assert octag._xobject.y_vertices[5] == -0.5
    xo.assert_allclose(octag._xobject.x_vertices[5], -0.5/np.tan(1.), atol=1e-14, rtol=0)


    assert octag._xobject.y_vertices[6] == -0.5
    xo.assert_allclose(octag._xobject.x_vertices[6], 0.5/np.tan(1.), atol=1e-14, rtol=0)
    assert octag._xobject.x_vertices[7] == 0.4
    xo.assert_allclose(octag._xobject.y_vertices[7], -0.4*np.tan(0.5), atol=1e-14, rtol=0)

    assert isinstance(line[line['m_polygon'].name_associated_aperture], xt.LimitPolygon)
    polyg = apertures[6]
    assert polyg.__class__.__name__ == 'LimitPolygon'
    assert len(polyg._xobject.x_vertices) == 3
    assert len(polyg._xobject.y_vertices) == 3
    # The below assertions don't match test_apertures.py::test_mad_import
    # TODO: Figure out what mad_loader.py:971 is doing? It seems wrong
    assert np.all(polyg.x_vertices == [5.8e-2, 5.8e-2, -8.8e-2])
    assert np.all(polyg.y_vertices == [3.5e-2, -3.5e-2, 0.0])


def test_aperture_setting():
    sequence = """
    m_ellipse: marker, apertype="ellipse", aperture={.2, .1};
    m_aper: marker, apertype="rectangle", aperture={.3, .4};
    
    line: sequence, l=1;
        m_ellipse, at=0;
        m_aper, at=0.1;
    endsequence;
    
    m_ellipse, aperture={.3, .2};  ! no apertype
    m_aper, apertype="ellipse", aperture={.5, .6};  ! change apertype
    """

    env = xt.load_madx_lattice(string=sequence)
    line = env.line

    assert line['m_ellipse_aper'].a == .3
    assert line['m_ellipse_aper'].b == .2

    assert line['m_aper_aper'].a == .5
    assert line['m_aper_aper'].b == .6


def test_import_thick_with_apertures_and_slice():
    sequence = """
    k1=0.2;
    tilt=0.1;

    elm: sbend,
        k1:=k1,
        l=1,
        angle=0.1,
        tilt=0.2,
        apertype="rectellipse",
        aperture={0.1,0.2,0.11,0.22},
        aper_tol={0.1,0.2,0.3},
        aper_tilt:=tilt,
        aper_offset={0.2, 0.3};

    seq: sequence, l=1;
    elm: elm, at=0.5;
    endsequence;
    """

    env = xt.load_madx_lattice(string=sequence)
    line = env.seq

    def _assert_eq(a, b):
        xo.assert_allclose(a, b, atol=1e-16)

    _assert_eq(line[f'elm_aper'].rot_s_rad, 0.1)
    _assert_eq(line[f'elm_aper'].shift_x, 0.2)
    _assert_eq(line[f'elm_aper'].shift_y, 0.3)
    _assert_eq(line[f'elm_aper'].max_x, 0.1)
    _assert_eq(line[f'elm_aper'].max_y, 0.2)
    _assert_eq(line[f'elm_aper'].a_squ, 0.11 ** 2)
    _assert_eq(line[f'elm_aper'].b_squ, 0.22 ** 2)

    _assert_eq(line[f'elm'].rot_s_rad, 0.2)

    line.slice_thick_elements(slicing_strategies=[Strategy(Uniform(2))])

    assert np.all(line.get_table().rows['elm_entry':'elm_exit'].name == [
        'elm_entry',                    # entry marker
        'elm_aper..0',                  # entry edge aperture
        'elm..entry_map',               # entry edge (+transform)
        'drift_elm..0',                 # drift 0
        'elm_aper..1',                  # slice 1 aperture
        'elm..0',                       # slice 0 (+transform)
        'drift_elm..1',                 # drift 1
        'elm_aper..2',                  # slice 2 aperture
        'elm..1',                       # slice 2 (+transform)
        'drift_elm..2',                 # drift 2
        'elm_aper..3',                  # exit edge aperture
        'elm..exit_map',                # exit edge (+transform)
        'elm_exit',                     # exit marker
    ])

    line.build_tracker(compile=False) # To resolve parents

    for i in range(4):
        _assert_eq(line[f'elm_aper..{i}'].resolve(line).rot_s_rad, 0.1)
        _assert_eq(line[f'elm_aper..{i}'].resolve(line).shift_x, 0.2)
        _assert_eq(line[f'elm_aper..{i}'].resolve(line).shift_y, 0.3)
        _assert_eq(line[f'elm_aper..{i}'].resolve(line).max_x, 0.1)
        _assert_eq(line[f'elm_aper..{i}'].resolve(line).max_y, 0.2)
        _assert_eq(line[f'elm_aper..{i}'].resolve(line).a_squ, 0.11 ** 2)
        _assert_eq(line[f'elm_aper..{i}'].resolve(line).b_squ, 0.22 ** 2)

    for i in range(2):
        _assert_eq(line[f'elm..{i}']._parent.rot_s_rad, 0.2)


import xtrack as xt
from scipy.constants import c as clight
from scipy.constants import e as qe

import numpy as np

fname = 'fccee_z'; pc_gev = 45.6
# fname = 'fccee_t'; pc_gev = 182.5


line = xt.Line.from_json(fname + '_thick_with_sol.json')

line.vars['on_sol_ip.1'] = 0
tw_sol_off = line.twiss(method='4d')
line.vars['on_sol_ip.1'] = 1
tw_sol_on = line.twiss(method='4d')
tw_local = line.twiss(start='ip.7', end='ip.2', init_at='ip.1',
                      init=tw_sol_off)

opt_l = line.match(
    solve=False,
    method='4d', n_steps_max=30,
    start='pqc2le.4', end='ip.1', init=tw_sol_off, init_at=xt.START,
    vary=[
        xt.VaryList(['acb1h.l1', 'acb2h.l1','acb1v.l1', 'acb2v.l1'], step=1e-8, tag='corr_l'),
        xt.VaryList(['ks1.l1', 'ks2.l1', 'ks3.l1', 'ks0.l1'], step=1e-7, tag='skew_l'),
        xt.VaryList(['corr_k1.l1', 'corr_k2.l1', 'corr_k3.l1', 'corr_k0.l1'], step=1e-6, tag='normal_l'),
    ],
    targets=[
        xt.TargetSet(['x', 'y'], value=tw_sol_off, tol=1e-7, at='ip.1', tag='orbit'),
        xt.TargetSet(['px', 'py'], value=tw_sol_off, tol=1e-10, at='ip.1', tag='orbit'),
        xt.TargetRmatrix(
                    r13=0, r14=0, r23=0, r24=0, # Y-X block
                    r31=0, r32=0, r41=0, r42=0, # X-Y block,
                    start='pqc2le.4', end='ip.1', tol=1e-6, tag='coupl'),
        xt.Target('mux', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
        xt.Target('muy', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
        xt.Target('betx', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=1, tol=1e-5),
        xt.Target('bety', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=10, tol=1e-6),
        xt.Target('alfx', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),
        xt.Target('alfy', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),

    ]
)


for iter in range(2):
    # Orbit alone
    opt_l.disable_all_targets(); opt_l.disable_all_vary()
    opt_l.enable_targets(tag='orbit'); opt_l.enable_vary(tag='corr_l'); opt_l.solve()

    # Coupling alone
    opt_l.disable_all_targets(); opt_l.disable_all_vary()
    opt_l.enable_targets(tag='coupl'); opt_l.enable_vary(tag='skew_l'); opt_l.solve()

    # phase, beta and alpha alone
    opt_l.disable_all_targets(); opt_l.disable_all_vary()
    opt_l.enable_vary(tag='normal_l')
    opt_l.enable_targets(tag='mu_ip'); opt_l.solve()
    opt_l.enable_targets(tag='bet_ip'); opt_l.solve()
    opt_l.enable_targets(tag='alf_ip'); opt_l.solve()

# All together
opt_l.enable_all_targets()
opt_l.enable_all_vary()
opt_l.solve()


opt_r = line.match(
    solve=False,
    method='4d', n_steps_max=30,
    start='ip.1', end='pqc2re.1', init=tw_sol_off, init_at=xt.END,
    vary=[
        xt.VaryList(['acb1h.r1', 'acb2h.r1','acb1v.r1', 'acb2v.r1'], step=1e-8, tag='corr_r'),
        xt.VaryList(['ks1.r1', 'ks2.r1', 'ks3.r1', 'ks0.r1'], step=1e-7, tag='skew_r'),
        xt.VaryList(['corr_k1.r1', 'corr_k2.r1', 'corr_k3.r1', 'corr_k0.r1'], step=1e-6, tag='normal_r'),
    ],
    targets=[
        xt.TargetSet(['x', 'y'], value=tw_sol_off, tol=1e-7, at='ip.1', tag='orbit'),
        xt.TargetSet(['px', 'py'], value=tw_sol_off, tol=1e-10, at='ip.1', tag='orbit'),
        xt.TargetRmatrix(r13=0, r14=0, r23=0, r24=0, # Y-X block
                         r31=0, r32=0, r41=0, r42=0, # X-Y block,
                         start='ip.1', end='pqc2re.1', tol=1e-6, tag='coupl'),
        xt.Target('mux', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
        xt.Target('muy', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
        xt.Target('betx', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=1, tol=1e-5),
        xt.Target('bety', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=10, tol=1e-6),
        xt.Target('alfx', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),
        xt.Target('alfy', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),

    ]
)

for iter in range(2):
    # Orbit alone
    opt_r.disable_all_targets(); opt_r.disable_all_vary()
    opt_r.enable_targets(tag='orbit'); opt_r.enable_vary(tag='corr_r'); opt_r.solve()

    # Coupling alone
    opt_r.disable_all_targets(); opt_r.disable_all_vary()
    opt_r.enable_targets(tag='coupl'); opt_r.enable_vary(tag='skew_r'); opt_r.solve()

    # phase, beta and alpha alone
    opt_r.disable_all_targets(); opt_r.disable_all_vary()
    opt_r.enable_vary(tag='normal_r')
    opt_r.enable_targets(tag='mu_ip'); opt_r.solve()
    opt_r.enable_targets(tag='bet_ip'); opt_r.solve()
    opt_r.enable_targets(tag='alf_ip'); opt_r.solve()

# All together
opt_r.enable_all_targets()
opt_r.enable_all_vary()
opt_r.solve()

tw_local_corr = line.twiss(start='ip.4', end='_end_point', init_at='ip.1',
                            init=tw_sol_off)
line.to_json(fname + '_thick_with_sol_corrected.json')

tw_sol_on_corrected = line.twiss(method='4d')

assert_allclose = np.testing.assert_allclose

# Check that tilt is present
assert_allclose(tw_sol_off['kin_xprime', 'ip.1'], np.tan(0.015), atol=1e-14, rtol=0)

# Check that solenoid introduces coupling
assert tw_sol_on.c_minus > 1e-4

# Check correction
tw_chk = tw_sol_on_corrected

assert_allclose(tw_chk['x', 'ip.1'], 0, atol=1e-8, rtol=0)
assert_allclose(tw_chk['y', 'ip.1'], 0, atol=1e-10, rtol=0)
assert_allclose(tw_chk['kin_xprime', 'ip.1'], tw_sol_off['kin_xprime', 'ip.1'],  atol=1e-9, rtol=0)
assert_allclose(tw_chk['kin_yprime', 'ip.1'], 0,  atol=1e-8, rtol=0)
assert_allclose(tw_chk['x', 'pqc2re.1'], 0, atol=5e-8, rtol=0)
assert_allclose(tw_chk['y', 'pqc2re.1'], 0, atol=5e-8, rtol=0)
assert_allclose(tw_chk['kin_xprime', 'pqc2re.1'], 0, atol=1e-8, rtol=0)
assert_allclose(tw_chk['kin_yprime', 'pqc2re.1'], 0, atol=1e-8, rtol=0)
assert_allclose(tw_chk['x', 'pqc2le.4'], 0, atol=5e-8, rtol=0)
assert_allclose(tw_chk['y', 'pqc2le.4'], 0, atol=5e-8, rtol=0)
assert_allclose(tw_chk['kin_xprime', 'pqc2le.4'], 0, atol=1e-8, rtol=0)
assert_allclose(tw_chk['kin_yprime', 'pqc2le.4'], 0, atol=1e-8, rtol=0)

assert_allclose(tw_chk['betx', 'ip.1'], tw_sol_off['betx', 'ip.1'], atol=0, rtol=5e-4)
assert_allclose(tw_chk['bety', 'ip.1'], tw_sol_off['bety', 'ip.1'], atol=0, rtol=5e-4)
assert_allclose(tw_chk['alfx', 'ip.1'], tw_sol_off['alfx', 'ip.1'], atol=1e-4, rtol=0)
assert_allclose(tw_chk['alfy', 'ip.1'], tw_sol_off['alfy', 'ip.1'], atol=1e-4, rtol=0)
assert_allclose(tw_chk['mux', 'ip.1'], tw_sol_off['mux', 'ip.1'], atol=1e-5, rtol=0)
assert_allclose(tw_chk['muy', 'ip.1'], tw_sol_off['muy', 'ip.1'], atol=1e-5, rtol=0)

assert_allclose(tw_chk['betx', 'pqc2re.1'], tw_sol_off['betx', 'pqc2re.1'], atol=0, rtol=5e-4)
assert_allclose(tw_chk['bety', 'pqc2re.1'], tw_sol_off['bety', 'pqc2re.1'], atol=0, rtol=5e-4)
assert_allclose(tw_chk['alfx', 'pqc2re.1'], tw_sol_off['alfx', 'pqc2re.1'], atol=2e-4, rtol=5e-4)
assert_allclose(tw_chk['alfy', 'pqc2re.1'], tw_sol_off['alfy', 'pqc2re.1'], atol=2e-4, rtol=5e-4)
assert_allclose(tw_chk['mux', 'pqc2re.1'], tw_sol_off['mux', 'pqc2re.1'], atol=1e-5, rtol=5e-5)
assert_allclose(tw_chk['muy', 'pqc2re.1'], tw_sol_off['muy', 'pqc2re.1'], atol=1e-5, rtol=5e-5)

assert_allclose(tw_chk['betx', 'pqc2le.4'], tw_sol_off['betx', 'pqc2le.4'], atol=0, rtol=5e-4)
assert_allclose(tw_chk['bety', 'pqc2le.4'], tw_sol_off['bety', 'pqc2le.4'], atol=0, rtol=5e-4)
assert_allclose(tw_chk['alfx', 'pqc2le.4'], tw_sol_off['alfx', 'pqc2le.4'], atol=2e-4, rtol=5e-4)
assert_allclose(tw_chk['alfy', 'pqc2le.4'], tw_sol_off['alfy', 'pqc2le.4'], atol=2e-4, rtol=5e-4)
assert_allclose(tw_chk['mux', 'pqc2le.4'], tw_sol_off['mux', 'pqc2le.4'], atol=1e-5, rtol=5e-5)
assert_allclose(tw_chk['muy', 'pqc2le.4'], tw_sol_off['muy', 'pqc2le.4'], atol=1e-5, rtol=5e-5)

assert tw_chk.c_minus < 1e-6
assert_allclose(tw_chk['betx2', 'ip.1'] / tw_chk['betx', 'ip.1'], 0, atol=1e-9)
assert_allclose(tw_chk['bety1', 'ip.1'] / tw_chk['bety', 'ip.1'], 0, atol=1e-9)
assert_allclose(tw_chk['betx2', 'pqc2re.1'] / tw_chk['betx', 'pqc2re.1'], 0, atol=1e-9)
assert_allclose(tw_chk['bety1', 'pqc2re.1'] / tw_chk['bety', 'pqc2re.1'], 0, atol=1e-9)
assert_allclose(tw_chk['betx2', 'pqc2le.4'] / tw_chk['betx', 'pqc2le.4'], 0, atol=1e-9)
assert_allclose(tw_chk['bety1', 'pqc2le.4'] / tw_chk['bety', 'pqc2le.4'], 0, atol=1e-9)

tt = line.get_table(attr=True)
s_ip = tt['s', 'ip.1']
P0_J = line.particle_ref.p0c[0] * qe / clight
brho = P0_J / qe / line.particle_ref.q0
Bz = tt.ks * brho

line_rad = line.copy()
line_rad.build_tracker()
line_rad.configure_radiation(model='mean')
line_rad.vars['voltca1'] = line.vv['voltca1_ref']
line_rad.vars['voltca2'] = line.vv['voltca2_ref']
line_rad.compensate_radiation_energy_loss()
tw_sol_on_corr_rad = line_rad.twiss()
mask_len = tt.length > 0
dE = -(np.diff(tw_sol_on_corr_rad.ptau) * tw_sol_on_corr_rad.particle_on_co.energy0[0])
dE_ds = tt.s * 0
dE_ds[mask_len] = dE[mask_len[:-1]] / tt.length[mask_len]

# plot
import matplotlib.pyplot as plt
plt.close('all')

plt.figure(5)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw_sol_on.s, tw_sol_on.betx2, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.betx2, label='correction on')
plt.ylabel(r'$\beta_{x,2}$ [m]')
plt.legend()

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s, tw_sol_on.bety1, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.bety1, label='correction on')
plt.ylabel(r'$\beta_{y,1}$ [m]')

plt.figure(6)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw_sol_on.s, tw_sol_off.kin_xprime, label='solenoid off')
plt.plot(tw_sol_on.s, tw_sol_on.kin_xprime, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.kin_xprime, label='correction on')
plt.ylabel("x'")
plt.legend()

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s, tw_sol_off.kin_yprime, label='solenoid off')
plt.plot(tw_sol_on.s, tw_sol_on.kin_yprime, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.kin_yprime, label='correction on')
plt.ylabel("y'")

plt.figure(7)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw_sol_on.s, tw_sol_on.x - tw_sol_off.x, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.x  - tw_sol_off.x, label='correction on')
plt.ylabel("x")
plt.legend()

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s, tw_sol_on.y - tw_sol_off.y, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.y  - tw_sol_off.y, label='correction on')
plt.ylabel("y")
plt.suptitle('Solenoid tilt is subtracted')

plt.figure(8)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw_sol_on.s, tw_sol_on.kin_xprime - tw_sol_off.kin_xprime, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.kin_xprime  - tw_sol_off.kin_xprime, label='correction on')
plt.ylabel("x'")
plt.legend()

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s, tw_sol_on.kin_yprime - tw_sol_off.kin_yprime, label='correction off')
plt.plot(tw_sol_on.s, tw_sol_on_corrected.kin_yprime  - tw_sol_off.kin_yprime, label='correction on')
plt.ylabel("y'")
plt.suptitle('Solenoid tilt is subtracted')

plt.figure(70, figsize=(6.4, 4.8*1.5))

ax1 = plt.subplot(3, 1, 1)
mask_ip = tt.name == 'ip.1'
plt.plot(tt.s[~mask_ip] - s_ip, Bz[~mask_ip])
plt.ylabel('Bz [T]')
plt.grid()

ax2 = plt.subplot(3, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s - s_ip, tw_sol_on_corrected.x  - tw_sol_off.x, label='correction on')
plt.ylabel("x [m]")
plt.grid()

ax3 = plt.subplot(3, 1, 3, sharex=ax1)
plt.plot(tw_sol_on.s - s_ip, tw_sol_on_corrected.y  - tw_sol_off.y, label='correction on')
plt.ylabel("y [m]")
plt.suptitle('Closed orbit - Solenoid tilt is subtracted')
plt.subplots_adjust(hspace=0.3, top=0.9)
plt.xlim(-5, 5)
plt.xlabel('s [m]')
plt.grid()

for nn in tt.rows['mcb.*'].name:
    for ax in [ax2, ax3]:
        ax.axvline(tt['s', nn] - s_ip, color='k', linestyle='--', alpha=0.3)

plt.figure(80, figsize=(6.4, 4.8*1.5))
s_ip = tt['s', 'ip.1']

ax1 = plt.subplot(3, 1, 1, sharex=ax1)
mask_ip = tt.name == 'ip.1'
plt.plot(tt.s[~mask_ip] - s_ip, Bz[~mask_ip])
plt.ylabel('Bz [T]')
plt.grid()

ax2 = plt.subplot(3, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s - s_ip, tw_sol_on_corrected.kin_xprime  - tw_sol_off.kin_xprime)
plt.ylabel("x'")
plt.grid()

ax3 = plt.subplot(3, 1, 3, sharex=ax1)
plt.plot(tw_sol_on.s - s_ip, tw_sol_on_corrected.kin_yprime  - tw_sol_off.kin_yprime)
plt.ylabel("y'")
plt.suptitle('Angles - Solenoid tilt is subtracted')
plt.subplots_adjust(hspace=0.3, top=0.9)
plt.xlim(-5, 5)
plt.xlabel('s [m]')
plt.grid()

for nn in tt.rows['mcb.*'].name:
    for ax in [ax2, ax3]:
        ax.axvline(tt['s', nn] - s_ip, color='k', linestyle='--', alpha=0.3)

plt.figure(90, figsize=(6.4, 4.8))
ax1 = plt.subplot(2, 1, 1, sharex=ax1)
plt.plot(tt.s[~mask_ip] - s_ip, Bz[~mask_ip])
plt.ylabel('Bz [T]')
plt.grid()

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s - s_ip, dE_ds * 1e-2 * 1e-3, '-', label='dE/ds')
plt.ylabel('dE/ds [keV/cm]')
plt.xlim(-5, 5)
plt.xlabel('s [m]')
plt.grid()
plt.suptitle('Radiated power')

plt.figure(85, figsize=(6.4, 4.8*1.5))
s_ip = tt['s', 'ip.1']

ax1 = plt.subplot(3, 1, 1)
mask_ip = tt.name == 'ip.1'
plt.plot(tt.s[~mask_ip] - s_ip, Bz[~mask_ip])
plt.ylabel('Bz [T]')
plt.grid()

ax2 = plt.subplot(3, 1, 2, sharex=ax1)
plt.plot(tw_sol_on.s - s_ip, tw_sol_on_corrected.bety1)
plt.ylabel(r"$\beta_{y,1}$")
plt.grid()

ax3 = plt.subplot(3, 1, 3, sharex=ax1)
plt.plot(tw_sol_on.s - s_ip, tw_sol_on_corrected.betx2)
plt.ylabel(r"$\beta_{x,2}$")
plt.suptitle('Local coupling')
plt.subplots_adjust(hspace=0.3, top=0.9)
plt.xlim(-10, 10)
plt.xlabel('s [m]')
plt.grid()
plt.suptitle('Local coupling')


plt.show()


import xtrack as xt
import numpy as np


k1 = 2.
length = 0.1

quad = xt.Quadrupole(k1=k1, length=length, shift_y=1e-3)

assert quad.shift_x == 0
assert quad.shift_y == 1e-3
assert quad._sin_rot_s == 0.0
assert quad._cos_rot_s == 1.0

p = xt.Particles(y=0, p0c=1e12)
quad.track(p)

assert_allclose = np.testing.assert_allclose
assert_allclose(p.py, k1 * length * -1e-3, rtol=5e-3, atol=0)

# Change the shift
quad.shift_y = 2e-3
p = xt.Particles(y=0, p0c=1e12)
quad.track(p)
assert_allclose(p.py, k1 * length * -2e-3, rtol=5e-3, atol=0)

# Make a line
line = xt.Line(elements=[quad])

# Slice the line:
line.slice_thick_elements(
    slicing_strategies=[xt.Strategy(xt.Uniform(3))])
line.build_tracker()

tt = line.get_table()
assert len(tt.rows['e0\.\..*']) == 3

p = xt.Particles(y=0, p0c=1e12)
line.track(p)

assert_allclose(p.py, k1 * length * -2e-3, rtol=5e-3, atol=0)

# Change the shift
quad.shift_y = 3e-3
p = xt.Particles(x=0, p0c=1e12)
line.track(p)

assert_allclose(p.py, k1 * length * -3e-3, rtol=5e-3, atol=0)


import numpy as np
import xtrack as xt

line = xt.Line.from_json('psb_04_with_chicane_corrected_thin.json')
line.build_tracker()

line.vars['on_chicane_k0'] = 1
line.vars['on_chicane_k2'] = 1
line.vars['on_chicane_beta_corr'] = 1
line.vars['on_chicane_tune_corr'] = 1

# Install monitor at foil
monitor = xt.ParticlesMonitor(start_at_turn=0, stop_at_turn=6000, num_particles=1)
line.discard_tracker()
line.insert_element(index='bi1.tstr1l1', element=monitor, name='monitor_at_foil')
line.build_tracker()

p = line.build_particles(x=0, px=0, y=0, py=0, delta=0, zeta=0)

line.enable_time_dependent_vars = True
line.dt_update_time_dependent_vars = 3e-6

print('Tracking...')
line.track(p, num_turns=6000, time=True)
print(f'Done in {line.time_last_track:.4} s')

import numpy as np
assert_allclose = np.testing.assert_allclose
assert_allclose(monitor.x[0, 0], -0.045936, rtol=0, atol=1e-5)
assert_allclose(monitor.x[0, 300], -0.04522354, rtol=0, atol=1e-5)
assert_allclose(monitor.x[0, 2500], -0.02256763, rtol=0, atol=1e-5)
assert_allclose(monitor.x[0, 4500], -0.00143883, rtol=0, atol=1e-5)
assert_allclose(monitor.x[0, 5500], 0.0, rtol=0, atol=1e-5)


import matplotlib.pyplot as plt

plt.close('all')
plt.figure(1)
plt.plot(monitor.x.T)

plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_collective_tracker_indices_one_turn(test_context):
    line = xt.Line(elements=[xt.Drift(length=1) for i in range(8)])

    line.elements[2].iscollective = True
    line.elements[2].move(_context=test_context)

    line.elements[5].iscollective = True
    line.elements[5].move(_context=test_context)

    line.build_tracker(_context=test_context)
    line.reset_s_at_end_turn = False

    particles = xp.Particles(x=0, px=0, _context=test_context)

    line.track(particles)

    assert particles.s[0] == len(line.elements)
    assert particles.at_turn[0] == 1

@for_all_test_contexts
def test_get_non_collective_line(test_context):

    line = xt.Line(
        elements=[xt.Drift(length=1) for i in range(8)],
        element_names=[f'e{i}' for i in range(8)]
    )
    line['e3'].iscollective = True
    e3_buffer = line['e3']._buffer
    e3 = line.get('e3')

    try:
        line.iscollective
    except RuntimeError:
        pass
    else:
        raise ValueError('This should have failed')

    try:
        line._buffer
    except RuntimeError:
        pass
    else:
        raise ValueError('This should have failed')

    line.build_tracker(_context=test_context)

    assert line.iscollective == True
    assert line['e0']._buffer is line._buffer
    assert line['e7']._buffer is line._buffer
    assert line['e3']._buffer is not line._buffer
    assert line['e3']._buffer is e3_buffer
    assert line.get('e3') is e3
    assert line.tracker.line is line

    nc_line = line._get_non_collective_line()

    # Check that the original line is untouched
    assert line.iscollective == True
    assert line['e0']._buffer is line._buffer
    assert line['e7']._buffer is line._buffer
    assert line['e3']._buffer is not line._buffer
    assert line['e3']._buffer is e3_buffer
    assert line.get('e3') is e3
    assert line.tracker.line is line

    assert nc_line.iscollective == False
    assert nc_line._buffer is line._buffer
    assert nc_line['e0']._buffer is line._buffer
    assert nc_line['e7']._buffer is line._buffer
    assert nc_line['e3']._buffer is line._buffer
    assert nc_line.get('e3') is not e3
    assert nc_line.get('e0') is line.get('e0')
    assert nc_line.get('e7') is line.get('e7')
    assert nc_line.tracker.line is nc_line

    xo.assert_allclose(nc_line.get_s_elements(), line.get_s_elements(),
                    rtol=0, atol=1e-15)

    assert nc_line.tracker is not line.tracker
    assert nc_line.tracker._tracker_data_cache is line.tracker._tracker_data_cache
    assert line.tracker._track_kernel is nc_line.tracker._track_kernel


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt

#################################
# Load a line and build tracker #
#################################

line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(
                    mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

particle_on_co = line.particle_ref.copy()
particle_on_co.x = 1e-4
particle_on_co.px = 0.
particle_on_co.y = 0.
particle_on_co.py = 0.
particle_on_co.zeta = 0.
particle_on_co.ptau = 0.

init = xt.twiss.TwissInit(
    particle_on_co=particle_on_co,
    W_matrix=np.eye(6),
    element_name=line.element_names[0],
)

line.twiss(init=init, start=line.element_names[0], end=line.element_names[10000])


# copyright ################################# #
# This file is part of the Xtrack Package.    #
# Copyright (c) CERN, 2024.                   #
# ########################################### #
import os

import cffi
import pytest

import xobjects as xo
import xtrack as xt


@pytest.fixture
def with_verbose():
    old_verbose = os.environ.get('XSUITE_VERBOSE', None)
    os.environ['XSUITE_VERBOSE'] = '1'

    yield

    if old_verbose is None:
        del os.environ['XSUITE_VERBOSE']
    else:
        os.environ['XSUITE_VERBOSE'] = old_verbose


def test_prebuild_kernels(mocker, tmp_path, temp_context_default_func, capsys, with_verbose):
    # Set up the temporary kernels directory
    kernel_defs = [
        ("111_test_module", {
            "config": {
                "XTRACK_MULTIPOLE_NO_SYNRAD": True,
                "XTRACK_GLOBAL_XY_LIMIT": 1.0,
                "XFIELDS_BB3D_NO_BEAMSTR": True,
                "XFIELDS_BB3D_NO_BHABHA": True,
            },
            "classes": [
                xt.Drift,
                xt.Cavity,
                xt.XYShift,
            ],
        }),
        ("000_test_module", {
            "config": {
                "XTRACK_MULTIPOLE_NO_SYNRAD": True,
                "XTRACK_GLOBAL_XY_LIMIT": 1.0,
                "XFIELDS_BB3D_NO_BEAMSTR": True,
                "XFIELDS_BB3D_NO_BHABHA": True,
            },
            "classes": [
                xt.Drift,
                xt.Cavity,
                xt.XYShift,
                xt.Bend,
            ],
        }),
    ]

    # Override the definitions with the temporary ones
    mocker.patch('xsuite.kernel_definitions.kernel_definitions', kernel_defs)
    mocker.patch('xsuite.prebuild_kernels.kernel_definitions', kernel_defs)
    # We need to change the default location so that loading the kernels works
    mocker.patch('xsuite.prebuild_kernels.XSK_PREBUILT_KERNELS_LOCATION',
                 tmp_path)
    mocker.patch('xsuite.XSK_PREBUILT_KERNELS_LOCATION',
                 tmp_path)

    # Try regenerating the kernels
    from xsuite.prebuild_kernels import regenerate_kernels
    regenerate_kernels(kernels=['111_test_module', '000_test_module'], location=tmp_path)

    # Check if the expected files were created
    so_file0, = tmp_path.glob('000_test_module.*.so')
    assert so_file0.exists()
    assert (tmp_path / '000_test_module.c').exists()
    assert (tmp_path / '000_test_module.json').exists()

    so_file1, = tmp_path.glob('111_test_module.*.so')
    assert so_file1.exists()
    assert (tmp_path / '111_test_module.c').exists()
    assert (tmp_path / '111_test_module.json').exists()

    # Test that reloading the kernel works, and it's the right one
    cffi_compile = mocker.patch.object(cffi.FFI, 'compile')

    line = xt.Line(elements=[xt.Drift(length=2.0)])

    # Build the tracker on a fresh context, so that the kernel comes from a file
    line.build_tracker(_context=xo.ContextCpu())

    p = xt.Particles(p0c=1e9, px=3e-6)
    line.track(p)

    assert p.x == 6e-6
    assert p.y == 0.0
    cffi_compile.assert_not_called()

    captured = capsys.readouterr()
    assert 'Found suitable prebuilt kernel `111_test_module`' in captured.out


def test_per_element_prebuild_kernels(mocker, tmp_path, temp_context_default_func):
    # Set up the temporary kernels directory
    kernel_defs = [
        ("test_module", {
            "config": {},
            "classes": [
                xt.Drift,
                xt.Cavity,
                xt.XYShift,
            ]
        }),
        ("test_module_rand", {
            "config": {},
            "classes": [],
            "extra_classes": [
                xt.RandomNormal,
            ]
        }),
    ]

    # Override the definitions with the temporary ones
    mocker.patch('xsuite.kernel_definitions.kernel_definitions', kernel_defs)
    mocker.patch('xsuite.prebuild_kernels.kernel_definitions', kernel_defs)
    # We need to change the default location so that loading the kernels works
    mocker.patch('xsuite.prebuild_kernels.XSK_PREBUILT_KERNELS_LOCATION',
                 tmp_path)
    mocker.patch('xsuite.XSK_PREBUILT_KERNELS_LOCATION',
                 tmp_path)

    # Try regenerating the kernels
    from xsuite.prebuild_kernels import regenerate_kernels
    regenerate_kernels(kernels=['test_module', 'test_module_rand'], location=tmp_path)

    # Check if the expected files were created
    so_file_exists = False
    for path in tmp_path.iterdir():
        if not path.name.startswith('test_module.'):
            continue
        if path.suffix not in ('.so', '.dll', '.dylib', '.pyd'):
            continue
        so_file_exists = True
    assert so_file_exists

    assert (tmp_path / 'test_module.c').exists()
    assert (tmp_path / 'test_module.json').exists()

    # Test that reloading the kernel works
    cffi_compile = mocker.patch.object(cffi.FFI, 'compile')

    drift = xt.Drift(length=2.0)

    p = xt.Particles(p0c=1e9, px=3e-6)
    drift.track(p)

    assert p.x == 6e-6
    assert p.y == 0.0

    rng = xt.RandomNormal()
    n_samples = 100
    samples = rng.generate(n_samples=n_samples, n_seeds=n_samples)
    assert len(samples) == n_samples

    cffi_compile.assert_not_called()



import xtrack as xt
import xpart as xp
from xpart.longitudinal.rf_bucket import RFBucket

import numpy as np
from scipy.constants import c as clight
from scipy.constants import e as qe
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt

gamma0 = 3. # defines the energy of the beam
gamma_transition = 4.
momentum_compaction_factor = 1 / gamma_transition**2
compensate_phase = True

particle_ref = xt.Particles(gamma0=gamma0,
                            mass0=xt.PROTON_MASS_EV)

circumference = 1000.
t_rev = circumference / (particle_ref.beta0[0] * clight)
f_rev = 1 / t_rev

energy_ref_increment =  50e3

eta = momentum_compaction_factor - 1 / particle_ref.gamma0[0]**2

h_rf = 40

f_rf = h_rf * f_rev
v_rf = 100e3
lag_rf = 180. if eta > 0. else 0.

# Compute momentum increment using auxiliary particle
dp0c_eV = energy_ref_increment / particle_ref.beta0[0]

if compensate_phase:
    phi_below = np.arcsin(dp0c_eV * particle_ref.beta0[0] / v_rf)
    phi_above = np.pi - phi_below
    lag_rf_above = np.rad2deg(phi_above)
    lag_rf_below = np.rad2deg(phi_below)
    if eta > 0:
        lag_rf = lag_rf_above
    else:
        lag_rf = lag_rf_below

otm = xt.LineSegmentMap(
    betx=1., bety=1,
    qx=6.3, qy=6.4,
    momentum_compaction_factor=momentum_compaction_factor,
    longitudinal_mode="nonlinear",
    voltage_rf=v_rf,
    frequency_rf=f_rf,
    lag_rf=lag_rf,
    length=circumference,
    energy_ref_increment=energy_ref_increment
)

line = xt.Line(elements={'otm': otm}, particle_ref=particle_ref)

tw = line.twiss()

p, matcher = xp.generate_matched_gaussian_bunch(
    line=line,
    num_particles=10_000,
    nemitt_x=2.5e-6,
    nemitt_y=2.5e-6,
    sigma_z=5,
    return_matcher=True)


# Logger (log every ten turns)
num_turns = 50_000
log_every = 20
n_log = num_turns // log_every
mon = xt.ParticlesMonitor(
    start_at_turn=0,
    stop_at_turn=1,
    n_repetitions=n_log,
    repetition_period=log_every,
    num_particles=len(p.x))

#########
# Track #
#########

jumped = False
i_jumped = None
while p.at_turn[0] < num_turns:
    print(f'Turn {p.at_turn[0]}/{num_turns}            ', end='\r', flush=True)

    # Phase jump
    if p.gamma0[0] > gamma_transition and not jumped:
        print(f'Jumped at turn: {p.at_turn[0]}')
        line['otm'].lag_rf = lag_rf_above
        i_jumped = p.at_turn[0]
        jumped = True

    # Track
    line.track(p, num_turns=100, turn_by_turn_monitor=mon)
    p.reorganize() # (put lost particles at the end)

############
# Plotting #
############

# Generate separatrix for video
i_separatrix = []
z_separatrix = []
delta_separatrix = []
for i_sep in np.arange(0, num_turns, 100):
    print(f'Separatrix {i_sep}/{num_turns}            ', end='\r', flush=True)
    line.particle_ref.gamma0 = mon.gamma0[i_sep//log_every, 0, 0]
    line['otm'].lag_rf = lag_rf_above if i_sep >= i_jumped else lag_rf_below
    try:
        rfb = line._get_bucket()
        i_separatrix.append(i_sep)
        z_separatrix.append(np.linspace(rfb.z_left, rfb.z_right, 1000))
        delta_separatrix.append(rfb.separatrix(z_separatrix[-1]))
    except:
        i_separatrix.append(i_sep)
        z_separatrix.append(np.zeros(1000))
        delta_separatrix.append(np.zeros(1000))
        z_separatrix[-1][:] = -1e20
        delta_separatrix[-1][:] = 0

sigma_z_rms = np.squeeze(mon.zeta.std(axis=1))
z_separatrix = np.array(z_separatrix)
delta_separatrix = np.array(delta_separatrix)
i_separatrix = np.array(i_separatrix)

plt.close('all')
fig1 = plt.figure(1)
plt.plot(mon.at_turn[:, 0, 0], sigma_z_rms)
plt.xlabel('Turn')
plt.ylabel('Bunch length [m]')
plt.show()

f_sep_z = interp1d(i_separatrix, z_separatrix, axis=0,
                   bounds_error=False, fill_value='extrapolate')
f_sep_delta = interp1d(i_separatrix, delta_separatrix, axis=0,
                       bounds_error=False, fill_value='extrapolate')

# Make movie (needed `conda install -c conda-forge ffmpeg``)
def update_plot(i_log, fig):
    i_turn = mon.at_turn[i_log, 0, 0]
    z_sep = f_sep_z(i_turn)
    delta_sep = f_sep_delta(i_turn)

    plot_separatrix = True
    if np.abs(i_turn - i_jumped) < 100:
        plot_separatrix = False

    phi_rf_deg = np.rad2deg(phi_above) if i_turn >= i_jumped else np.rad2deg(phi_below)
    plt.clf()
    plt.plot(mon.zeta[i_log, :], mon.delta[i_log, :], '.', markersize=1)
    if plot_separatrix:
        plt.plot(z_sep, delta_sep, color='C1', linewidth=2)
        plt.plot(z_sep, -delta_sep, color='C1', linewidth=2)
    plt.xlim(-10, 10)
    plt.ylim(-10e-3, 10e-3)
    plt.xlabel('z [m]')
    plt.ylabel(r'$\Delta p / p_0$')
    plt.title(f'Turn {i_turn} '
              r'$\sigma_\zeta = $' f'{sigma_z_rms[i_log]:.2f}\n'
              r'$\gamma_0 = $' f'{mon.gamma0[i_log, 0, 0]:.2f} '
              r'$\gamma_t = $' f'{gamma_transition:.2f} '
              r'$\phi_{\mathrm{rf}} = $' f'{phi_rf_deg:.2f}')
    plt.subplots_adjust(left=0.2, top=0.82)
    plt.grid(alpha=0.5)

fig = plt.figure()
from matplotlib.animation import FFMpegFileWriter
moviewriter = FFMpegFileWriter(fps=15)
with moviewriter.saving(fig, 'transition.mp4', dpi=100):
    for j in range(0, len(mon.zeta[:, 0, 0]), 1):
        print(f'Frame {j}/{len(mon.zeta[:, 0, 0])}            ', end='\r', flush=True)
        update_plot(j, fig)
        moviewriter.grab_frame()



import xtrack as xt
from cpymad.madx import Madx

# Load a very simple sequence from MAD-X
mad = Madx()
mad.input("""
    seq: sequence, l=4;
    b1: sbend, at=0.5, angle=0.2, l=1;
    b2: sbend, at=2.5, angle=0.3, l=1;
    endsequence;

    beam;
    use,sequence=seq;
""")

line = xt.Line.from_madx_sequence(mad.sequence.seq)
line.build_tracker()

print('The line as imported from MAD-X:')
line.get_table().show()

# Shift and tilt selected elements
line['b1'].shift_x = -0.01
line['b1'].rot_s_rad = 0.8
line['b2'].shift_s = 0.02
line['b2'].rot_s_rad = -0.8

tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'shift_x', 'shift_y', 'shift_s', 'rot_s_rad']
# returns:
#
# name       s element_type isthick shift_x shift_y shift_s rot_s_rad
# seq$start  0 Marker         False       0       0       0         0
# b1         0 Bend            True   -0.01       0       0       0.8
# drift_0    1 Drift           True       0       0       0         0
# b2         2 Bend            True       0       0    0.02      -0.8
# drift_1    3 Drift           True       0       0       0         0
# seq$end    4 Marker         False       0       0       0         0
# _end_point 4                False       0       0       0         0



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.vars['vrf400'] = 16
line.build_tracker()


# Twiss
tw = line.twiss()

# Inspect tunes and chromaticities
tw.qx # Horizontal tune
tw.qy # Vertical tune
tw.dqx # Horizontal chromaticity
tw.dqy # Vertical chromaticity

# Plot closed orbit and lattice functions
import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1, figsize=(6.4, 4.8*1.5))
spbet = plt.subplot(3,1,1)
spco = plt.subplot(3,1,2, sharex=spbet)
spdisp = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(tw.s, tw.betx)
spbet.plot(tw.s, tw.bety)
spbet.set_ylabel(r'$\beta_{x,y}$ [m]')

spco.plot(tw.s, tw.x)
spco.plot(tw.s, tw.y)
spco.set_ylabel(r'(Closed orbit)$_{x,y}$ [m]')

spdisp.plot(tw.s, tw.dx)
spdisp.plot(tw.s, tw.dy)
spdisp.set_ylabel(r'$D_{x,y}$ [m]')
spdisp.set_xlabel('s [m]')

fig1.suptitle(
    r'$q_x$ = ' f'{tw.qx:.5f}' r' $q_y$ = ' f'{tw.qy:.5f}' '\n'
    r"$Q'_x$ = " f'{tw.dqx:.2f}' r" $Q'_y$ = " f'{tw.dqy:.2f}'
    r' $\gamma_{tr}$ = '  f'{1/np.sqrt(tw.momentum_compaction_factor):.2f}'
)

fig1.subplots_adjust(left=.15, right=.92, hspace=.27)
plt.show()



import xtrack as xt
from pathlib import Path

line_path = (Path(xt.__file__).parent / '../test_data/fcc_ee/fccee_h_thin.json').absolute()

line = xt.Line.from_json(line_path)
line.build_tracker()

import time
start_time = time.time()
tw = line.twiss()
end_time = time.time()
print(f"Vanilla twiss: {end_time - start_time} seconds")

line.configure_radiation(model='mean')
line.build_tracker()
line.twiss()

start_time = time.time()
tw = line.twiss()
end_time = time.time()
print(f"With radiation twiss: {end_time - start_time} seconds")

line.twiss(eneloss_and_damping=True)
start_time = time.time()
tw = line.twiss(eneloss_and_damping=True)
end_time = time.time()
print(f"Energy loss and damping twiss: {end_time - start_time} seconds")
